__authors__ = ["Christian O'Reilly", "Pierre-Alexandre Fonta"]
__maintainer__ = "Pierre-Alexandre Fonta"

import getpass
import os
import pickle
import time
import webbrowser
from glob import glob
from os.path import join
from threading import Thread

import numpy as np
from PyQt5.QtCore import QUrl, pyqtSlot, pyqtSignal, QItemSelection, Qt, QModelIndex
from PyQt5.QtGui import QIcon, QDesktopServices
from PyQt5.QtWidgets import (QAction, QMainWindow, QLabel, QMessageBox, QDialog,
                             QTabWidget, QSplitter, QWidget, QVBoxLayout,
                             QGroupBox, QPushButton, QLineEdit, QGridLayout,
                             QTableView, QAbstractItemView, QListWidget,
                             QListWidgetItem, QApplication, QFileDialog)

from nat.annotation import Annotation
from nat.gitManager import GitManager, GitMngError
from nat.id import checkID
from nat.ontoManager import OntoManager
from nat.restClient import RESTClient, RESTImportPDFErr
from nat.tag import Tag
from nat.utils import Id2FileName  # , fileName2Id
from neurocurator.utils import package_directory
from neurocurator.zotero_widget import ZoteroTableWidget
from requests.exceptions import ConnectionError
from .addOntoTermDlg import AddOntoTermDlg
from .annotWidgets import EditAnnotWgt
from .annotationListModel import AnnotationListModel
from .autocomplete import AutoCompleteEdit
from .experimentalPropertyWgt import ExpPropWgt
from .modParamWidgets import ParamModWgt
from .searchInterface import SearchWgt
from .searchOntoWgt import OntoOnlineSearch
from .settingsDlg import getSettings, SettingsDlg
from .suggestedTagMng import TagSuggester
from .tagWidget import TagWidget
from .uiUtilities import errorMessage, disableTextWidget


class Window(QMainWindow):

    selectedAnnotationChangedConfirmed = pyqtSignal()
    annotationCleared                  = pyqtSignal()
    savingNeeded                       = pyqtSignal(bool)

    def popUpSettingsDlg(self):
        self.settings = getSettings(True)
        if self.settings is None:
            self.close()
            self.deleteLater()
            return
        else:
            # Refresh the git manage to ensure that it is representative of
            # the actual settings.
            try:
                self.gitMng = GitManager(self.settings.config["GIT"])
            except KeyError:
                self.popUpSettingsDlg()

    def __init__(self, parent=None):
        super().__init__(parent)

        # Annotation curently being displayed, modified, or created
        self.currentAnnotation = None

        # True when the current annotation has been modified and require saving
        self.needSaving         = False

        self.needSavingDisabled = False

        # FIXME Delayed refactoring. Use signals/slots to update the message.
        self._statusBar = self.statusBar()
        self.statusLabel = QLabel("")
        self._statusBar.addWidget(self.statusLabel)

        # Get the system username. It will be used to identify the curator
        # in the annotations.
        self.username = getpass.getuser()

        self.loadPersistTag()

        # Number of characters displayed before and after annotations when
        # the annotation is displayed with its context.
        self.contextLength = 30
        
        self.needPush = False

        # True when the annotationEdt field has been modified BY THE USER
        self.detectAnnotChange     = False
    
        # Load the ontological trees (pre-save for efficiency)
        self.builtOntoTrees()

        # Load the tag suggester (based on saved tagging history)
        self.tagSuggester = TagSuggester.load()

        # Load saved settings
        self.settings = getSettings()
        if self.settings is None:
            self.close()
            self.deleteLater()
            return

        # Load the object used to transparently interact with GIT to save annotations
        # using versioning.
        def getGitMng(cleanDirty=False):
            try:
                self.gitMng = GitManager(self.settings.config["GIT"], cleanDirty)
            except KeyError:
                self.popUpSettingsDlg()
            except GitMngError as e:
                msgBox = QMessageBox()
                msgBox.setStandardButtons(QMessageBox.Cancel)
                msgBox.setWindowTitle("GIT repository is dirty")
                msgBox.setText(str(e))
                button = msgBox.addButton("commit", QMessageBox.YesRole)
                msgBox.setDefaultButton(button)
                msgBox.exec_()
                if msgBox.clickedButton() == button:
                    getGitMng(cleanDirty=True)
                else:
                    raise
        getGitMng()

        # Setup the REST client
        try:
            self.restClient = RESTClient(self.settings.config["REST"]["serverURL"])
        except KeyError:
            self.popUpSettingsDlg()

        # Load from config the path where the GIT database is located.
        self.dbPath   = os.path.abspath(os.path.expanduser(self.settings.config["GIT"]["local"]))

        self.setupWindowsUI()
        # Must be called after the creation of the widgets to connect to their slots.
        self.setupMenus()

        self.firstShow = True

    @property
    def needSaving(self):
        return self.__needSaving

    @needSaving.setter
    def needSaving(self, needSaving):
        self.__needSaving = needSaving
        self.savingNeeded.emit(needSaving)



    def loadPersistTag(self):
        
        if os.path.isfile(os.path.join(os.path.dirname(__file__), 'persistTag.pkl')):
            try:
                with open(os.path.join(os.path.dirname(__file__), 'persistTag.pkl'), 'rb') as f:
                    self.selectedTagPersist, self.suggestTagPersist  = pickle.load(f)
            except:
                self.selectedTagPersist = {}
                self.suggestTagPersist  = []    
        else:
            self.selectedTagPersist = {}
            self.suggestTagPersist  = []

    def savePersistTag(self):
        with open(os.path.join(os.path.dirname(__file__), 'persistTag.pkl'), 'wb') as f:
            pickle.dump((self.selectedTagPersist, self.suggestTagPersist), f)



    def closeEvent(self, event):
        # FIXME Delayed refactoring. Do settings management with QSettings.
        window_settings = self.settings.config['WINDOW']
        window_settings['mainSplitterPos'] = str(self.mainWidget.sizes())
        window_settings['leftSplitterPos'] = str(self.leftPanel.sizes())
        window_settings['rightSplitterPos'] = str(self.rightPanel.sizes())
        window_settings['paramModWgtSplitterPos'] = str(self.modParamWgt.rootLayout.sizes())

        zotero_view = self.zotero_widget.view

        zotero_column_count = zotero_view.model().columnCount()
        colWidths = str([zotero_view.columnWidth(i) for i in range(zotero_column_count)])
        window_settings['zotTableViewColWidth'] = colWidths

        zotero_table_header = zotero_view.horizontalHeader()
        # FIXME Delayed refactoring. "If no section has a sort indicator, return value is undefined".
        zotero_sort_order = zotero_table_header.sortIndicatorOrder()
        zotero_sort_column = zotero_table_header.sortIndicatorSection()
        window_settings['zotTableSortOrder'] = str(int(zotero_sort_order))
        window_settings['zotTableSortCol'] = str(zotero_sort_column)

        colWidths = str([self.annotListTblWdg.columnWidth(i) for i in range(self.annotTableModel.columnCount())])
        window_settings['annotTableViewColWidth'] = colWidths
        window_settings['annotTableSortOrder']      = str(int(self.annotTableModel.sortOrder))
        window_settings['annotTableSortCol']      = str(self.annotTableModel.sortCol)

        self.settings.save()

        if self.needSaving:
            msgBox = QMessageBox(self)
            msgBox.setWindowTitle("Unsaved annotation")
            msgBox.setText("The current annotations has been modified. Do you want to save it before quitting?")
            msgBox.setStandardButtons(QMessageBox.No | QMessageBox.Yes)
            msgBox.setDefaultButton(QMessageBox.Yes)
            if msgBox.exec_() == QMessageBox.Yes:
                self.saveAnnotation()
            
        if self.needPush:
            msgBox = QMessageBox(self)
            msgBox.setWindowTitle("GIT Push recommanded")
            msgBox.setText("Some files have been modified. Do you want to push these modifications to the server before quitting?")
            msgBox.setStandardButtons(QMessageBox.No | QMessageBox.Yes)
            msgBox.setDefaultButton(QMessageBox.Yes)
            if msgBox.exec_() == QMessageBox.Yes:
                self.pushToServer()

        event.accept()



    def showEvent(self, event):
        # FIXME Delayed refactoring. Save/Restore also the main window size.
        if self.firstShow:
            if 'mainSplitterPos' in self.settings.config['WINDOW']:
                self.mainWidget.setSizes(eval(self.settings.config['WINDOW']['mainSplitterPos']))
            if 'leftSplitterPos' in self.settings.config['WINDOW']:
                self.leftPanel.setSizes(eval(self.settings.config['WINDOW']['leftSplitterPos']))
            if 'rightSplitterPos' in self.settings.config['WINDOW']:
                self.rightPanel.setSizes(eval(self.settings.config['WINDOW']['rightSplitterPos']))
            if 'paramModWgtSplitterPos' in self.settings.config['WINDOW']:
                self.modParamWgt.setRootLayoutSizes(eval(self.settings.config['WINDOW']['paramModWgtSplitterPos']))

            if 'zotTableViewColWidth' in self.settings.config['WINDOW']:            
                for i, width in enumerate(eval(self.settings.config['WINDOW']['zotTableViewColWidth'])):
                    self.zotero_widget.view.setColumnWidth(i, width)
            if 'annotTableViewColWidth' in self.settings.config['WINDOW']:
                for i, width in enumerate(eval(self.settings.config['WINDOW']['annotTableViewColWidth'])):
                    self.annotListTblWdg.setColumnWidth(i, width)

            if 'annotTableSortOrder' in self.settings.config['WINDOW']:
                self.annotTableModel.sortOrder = Qt.SortOrder(int(self.settings.config['WINDOW']['annotTableSortOrder']))

            if 'annotTableSortCol' in self.settings.config['WINDOW']:
                self.annotTableModel.sortCol = int(self.settings.config['WINDOW']['annotTableSortCol'])

            # FIXME Delayed refactoring. Do settings management with QSettings.
            window_settings = self.settings.config['WINDOW']
            is_zotero_sort_order_set = 'zotTableSortOrder' in window_settings
            is_zotero_sort_column_set = 'zotTableSortCol' in window_settings
            if is_zotero_sort_order_set and is_zotero_sort_column_set:
                zotero_sort_column = int(window_settings['zotTableSortCol'])
                zotero_sort_order = Qt.SortOrder(int(window_settings['zotTableSortOrder']))
                # Triggers a sorting.
                self.zotero_widget.view.horizontalHeader().setSortIndicator(zotero_sort_column, zotero_sort_order)

            self.annotListTblWdg.sortByColumn(self.annotTableModel.sortCol, 
                                           self.annotTableModel.sortOrder)

            self.firstShow = False

    def setupMenus(self):


        self.statusBar()

        menu_bar = self.menuBar()
        
        #######################################################################
        # FILE MENU
        exitAction = QAction(QIcon('exit.png'), '&Exit', self)
        exitAction.setShortcut('Ctrl+Q')
        exitAction.setStatusTip('Exit application')
        exitAction.triggered.connect(self.close)
        
        fileMenu = menu_bar.addMenu('&File')
        fileMenu.addAction(exitAction)        

        #######################################################################
        # EDIT MENU
        openPreferencesAction = QAction(QIcon(), '&Preferences', self)
        openPreferencesAction.setStatusTip('Edit preferences')
        openPreferencesAction.triggered.connect(self.editPreferences)

        editMenu = menu_bar.addMenu('&Edit')
        editMenu.addAction(openPreferencesAction)



        #######################################################################
        # COMMAND MENU
        refreshLocalOntoAction = QAction(QIcon(), 'Refresh local &ontologies', self)
        refreshLocalOntoAction.setStatusTip('Refresh local ontologies')
        refreshLocalOntoAction.triggered.connect(self.refreshLocalOnto)

        pushToServerAction = QAction(QIcon(), '&Push annotations to server', self)
        pushToServerAction.setStatusTip('Push modifications to the server')
        pushToServerAction.triggered.connect(self.pushToServer)

        addToOntologyAction = QAction(QIcon(), '&Add a term to ontologies', self)
        addToOntologyAction.setStatusTip('Add a new ontological term')
        addToOntologyAction.triggered.connect(self.addToOntology)

        addModParamTypeAction = QAction(QIcon(), '&Add a modeling parameter type', self)
        addModParamTypeAction.setStatusTip('Add a new type of modeling parameter')
        addModParamTypeAction.triggered.connect(self.addModParamType)
        

        commandMenu = menu_bar.addMenu('&Command')
        commandMenu.addAction(pushToServerAction)
        commandMenu.addAction(refreshLocalOntoAction)
        commandMenu.addAction(addToOntologyAction)
        commandMenu.addAction(addModParamTypeAction)

        # Zotero menu section.

        refresh_zotero_action = QAction("Refresh database", self)
        refresh_zotero_action.setStatusTip("Refresh the Zotero database")
        refresh_zotero_action.triggered.connect(self.zotero_widget.refresh_database)

        add_zotero_action = QAction("Add reference", self)
        add_zotero_action.setStatusTip("Add a reference to the Zotero database")
        add_zotero_action.triggered.connect(self.zotero_widget.add_reference)

        edit_zotero_action = QAction("Edit reference", self)
        edit_zotero_action.setStatusTip('Edit selected Zotero reference')
        edit_zotero_action.triggered.connect(self.zotero_widget.edit_reference)

        self.zotero_menu = menu_bar.addMenu("Zotero")
        self.zotero_menu.addAction(refresh_zotero_action)
        self.zotero_menu.addAction(add_zotero_action)
        self.zotero_menu.addAction(edit_zotero_action)

    @pyqtSlot()
    def zotero_refresh_started(self):
        self.zotero_menu.setDisabled(True)
        self.statusBar().showMessage("Refreshing the Zotero database...")

    @pyqtSlot()
    def zotero_refresh_finished(self):
        self.zotero_menu.setEnabled(True)
        self.statusBar().showMessage("The Zotero database has been refreshed.", 10 * 1000)

    def addToOntology(self):
        addToOntoDlg = AddOntoTermDlg(self)
        if addToOntoDlg.exec_() == QDialog.Accepted:
            pass


    def addModParamType(self):
        pass


    def refreshLocalOnto(self):
        self.builtOntoTrees(recompute=True)
        #TODO: Need to refresh auto-completion lists if we want it to take
        # this refreshed ontology into account without restarting the app.

    def setupWindowsUI(self) :
        self.setupPaperGB()
        self.setupListAnnotGB()
        self.setupEditAnnotGB()
        self.setupTagAnnotGB()
        self.modParamWgt = ParamModWgt(self)
        self.expPropWgt = ExpPropWgt(self)

        # Main layout
        # FIXME Delayed refactoring. Create a dedicated QTabWidget object.
        self.mainTabs = QTabWidget(self)

        # FIXME Delayed refactoring. Do settings management with QSettings.
        zotero_settings = self.settings.config["ZOTERO"]
        directory = package_directory()

        # NB: Don't specify a parent for widgets to be added to a QTabWidget.
        self.zotero_widget = ZoteroTableWidget(zotero_settings, directory, self.checkIdInDB, self.dbPath, self)

        self.zotero_widget.view.doubleClicked.connect(self.changeTagToAnnotations)

        # The signals currentRowChanged and selectionChanged of QItemSelectionModel
        # are emitted when QSortFilterProxyModel::setFilterFixedString() is called.
        # The signal currentRowChanged is emitted when the user clicks on the
        # application for the first time.
        selection_model = self.zotero_widget.view.selectionModel()  # Necessary.
        selection_model.selectionChanged.connect(self.paperSelectionChanged)

        self.zotero_widget.refresh_thread.started.connect(self.zotero_refresh_started)
        self.zotero_widget.refresh_thread.finished.connect(self.zotero_refresh_finished)

        self.mainTabs.addTab(self.zotero_widget, "References (Zotero)")

        self.taggingTabs = QTabWidget(self)
        self.taggingTabs.addTab(self.tagAnnotGroupBox, "Tagging")
        self.taggingTabs.addTab(self.modParamWgt,      "Parameters")
        #self.taggingTabs.addTab(self.expPropWgt,       "Relevant experimental properties")

        self.rightPanel = QSplitter(Qt.Vertical, self)
        self.rightPanel.setOrientation(Qt.Vertical)

        self.leftPanel = QSplitter(Qt.Vertical, self)
        self.leftPanel.setOrientation(Qt.Vertical)

        paperPannel = QWidget(self)
        paperPannel.setLayout(QVBoxLayout())
        paperPannel.layout().addWidget(self.paperGroupBox)
        paperPannel.layout().addWidget(self.listAnnotGroupBox)

        bottomPannel = QWidget(self)
        bottomPannel.setLayout(QVBoxLayout())
        bottomPannel.layout().addWidget(self.taggingTabs)



        self.rightPanel.addWidget(bottomPannel)


        self.leftPanel.addWidget(paperPannel)
        self.leftPanel.addWidget(self.editAnnotWgt)
        self.expPropGB = QGroupBox("Relevant experimental properties")
        expPropLayout = QVBoxLayout(self.expPropGB)
        expPropLayout.addWidget(self.expPropWgt)
        self.leftPanel.addWidget(self.expPropGB)

        self.mainWidget = QSplitter(Qt.Horizontal, self)
        self.mainWidget.addWidget(self.leftPanel)
        self.mainWidget.addWidget(self.rightPanel)

        self.mainTabs.addTab(self.mainWidget, "Annotations")    

        self.searchTabs =  QTabWidget(self)
        self.annotSearchWgt = SearchWgt("Annotation", self)
        self.paramSearchWgt = SearchWgt("Parameter", self)
        
        self.searchTabs.addTab(self.annotSearchWgt, "Annotations")
        self.searchTabs.addTab(self.paramSearchWgt, "Parameters")
        
        self.annotSearchWgt.annotationSelected.connect(self.viewAnnotation)
        self.paramSearchWgt.parameterSelected.connect(self.viewParameter)
        
        self.mainTabs.addTab(self.searchTabs, "Search")

        #self.mainWidget.setSizes([1,1])
        #self.mainWidget.setStretchFactor(0, 1)
        #self.mainWidget.setStretchFactor(1, 1)
        self.setCentralWidget(self.mainTabs)

        # Initial behavior
        self.taggingTabs.setDisabled(True)

    @pyqtSlot(QModelIndex)
    def changeTagToAnnotations(self, index):
        # FIXME Delayed refactoring. No use of the index sent by ZoteroTableView::doubleClicked?
        self.mainTabs.setCurrentIndex(1)
    
    def setupPaperGB(self):
        # Widgets
        self.openPDFBtn                = QPushButton('Open PDF', self)
        self.IdTxt                    = QLineEdit('', self)

        # Signals
        self.openPDFBtn.clicked.connect(self.openPDF)

        # Layout
        self.paperGroupBox = QGroupBox("Paper")
        gridPaper = QGridLayout(self.paperGroupBox)
        gridPaper.addWidget(QLabel('ID', self), 0, 0)
        gridPaper.addWidget(self.IdTxt, 0, 1)
        gridPaper.addWidget(self.openPDFBtn, 0, 3)
        #gridPaper.addWidget(self.refEdt, 1, 0, 1, 4)

        # Initial behavior
        self.paperGroupBox.setDisabled(True)
        #disableTextWidget(self.refEdt)
        disableTextWidget(self.IdTxt)



    def setupListAnnotGB(self):
        self.cancelledAnnotSelectinChange = False        

        # Widget        
        self.annotListTblWdg      = QTableView()
        self.annotTableModel     = AnnotationListModel(parent=self)
        self.annotListTblWdg.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.annotListTblWdg.setSelectionMode(QAbstractItemView.SingleSelection)
        self.annotListTblWdg.setModel(self.annotTableModel)

        # Signals
        self.annotationSelectionModel = self.annotListTblWdg.selectionModel()
        self.annotationSelectionModel.selectionChanged.connect(self.selectedAnnotationChanged)
        self.annotTableModel.layoutChanged.connect(self.annotTableLayoutChanged)


        # Layout
        self.listAnnotGroupBox     = QGroupBox("Listing of existing annotations")
        gridListAnnotations     = QGridLayout(self.listAnnotGroupBox)
        gridListAnnotations.addWidget(self.annotListTblWdg, 0, 0)
        
        # Initial behavior
        self.listAnnotGroupBox.setDisabled(True)        
        self.annotListTblWdg.setSortingEnabled(True)
        self.annotListTblWdg.horizontalHeader().sectionClicked.connect(self.setAnnotSortCol)


    def setAnnotSortCol(self, col):
        self.annotTableModel.sortCol = col        
        self.annotTableModel.sortOrder = self.annotListTblWdg.horizontalHeader().sortIndicatorOrder()


    def setupEditAnnotGB(self):

        self.editAnnotSubWgt = EditAnnotWgt(self)

        self.editAnnotWgt = QGroupBox("Annotation details")
        layout               = QVBoxLayout(self.editAnnotWgt)
        layout.addWidget(self.editAnnotSubWgt)
        self.editAnnotWgt.setEnabled(False)


    def setupTagAnnotGB(self):
        # Widgets        

        # This fields provide a text fields that can be used
        # to enter tags. It is using an autocompletion scheme which
        # suggests available ontological tags according to entrer word.
        # Matching between entered text and ontological concepts are
        # is not case-sensitive and is not using a prefix scheme (i.e., 
        # matching can be done anywhere within the strings, not only with
        # their begininings)  
        self.tagEdit = AutoCompleteEdit(self)
        self.tagEdit.setMinimumWidth(10)

        self.updateAutoCompleteTagList()

        # List tags that have been selected by the user
        self.selectedTagsWidget = QListWidget(self)
        self.selectedTagsWidget.showMaximized()
        self.selectedTagsWidget.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.selectedTagsWidget.setSelectionMode(QAbstractItemView.SingleSelection)

        # List tags that are suggested to the user based on his previous tagging
        # history and on tags that have already been used for other annotations 
        # on this paper.
        self.suggestedTagsWidget = QListWidget(self)
        self.suggestedTagsWidget.showMaximized()
        self.suggestedTagsWidget.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.suggestedTagsWidget.setSelectionMode(QAbstractItemView.SingleSelection)

        self.onlineOntoWgt     = OntoOnlineSearch(self)
        self.onlineOntoWgt.tagSelected.connect(self.ontoTagSelected)


        self.tagSelectionTabs = QTabWidget(self)
        self.tagSelectionTabs.addTab(self.suggestedTagsWidget, "Suggested tags")     
        self.tagSelectionTabs.addTab(self.onlineOntoWgt, "Search online ontologies")     

    
        # Signals
        self.tagEdit.activated[str].connect(self.tagSuggestionSelected)
        self.tagEdit.editTextChanged.connect(self.editTextChanged)

        # Layout
        self.tagAnnotGroupBox     = QGroupBox()
        gridTagAnnotations     = QGridLayout(self.tagAnnotGroupBox)
        gridTagAnnotations.addWidget(QLabel('Annotation tags', self), 0, 0)
        gridTagAnnotations.addWidget(self.selectedTagsWidget, 1, 0)
        gridTagAnnotations.addWidget(self.tagEdit, 2, 0)
        gridTagAnnotations.addWidget(self.tagSelectionTabs, 0, 1, 3, 1)
        gridTagAnnotations.setColumnStretch(0, 1)
        gridTagAnnotations.setColumnStretch(1, 1)




    def updateAutoCompleteTagList(self):
        # Sort list of suggestions so that more often used tags 
        # are on the top of the autocompletion list
        ids = list(self.tagSuggester.usedTag.keys())
        usage = list(self.tagSuggester.usedTag.values())
        ids.sort(key=dict(zip(ids, usage)).get, reverse=True)

        usedNames = [self.dicData[id] for id in ids if id in self.dicData] 


        allNames  = np.array(list(self.dicData.values()))
 
        # TODO: LINE ADDED TEMPORALLY. TO REMOVE. THERE SHOULD BE NO NONE VALUES
        #       HERE. ASSERTS HAS BEEN ADDED TO AVOID THIS TO HAPPEN AGAIN.
        allNames = np.array([name for name in allNames if not name is None])

        assert(not np.any([name is None for name in allNames]))

        # Putting used tag at the top of the list. 
        allNames = np.concatenate([usedNames, allNames[np.logical_not(np.in1d(allNames, usedNames))]])
        
        self.tagEdit.setModel(allNames)
    




    @pyqtSlot(str, str)
    def ontoTagSelected(self, term, curie):
        self.addTagToAnnotation(curie, term)

    @pyqtSlot(object)
    def viewAnnotation(self, annotation):
        # FIXME Delayed refactoring. Only one parameter is sent.

        zotero_view = self.zotero_widget.view
        zotero_model = zotero_view.model()

        # Deactivate the filtering of the Zotero QTableView.
        self.zotero_widget.filter_edit.clear()
        self.zotero_widget.filter_edit.clearFocus()

        match = zotero_model.match(zotero_model.index(0, 0), Qt.DisplayRole,
                                   annotation.pubId, 1, Qt.MatchExactly)
        if match:
            zotero_view.selectRow(match[0].row())
        else:
            print(annotation)
            raise ValueError("No matching annotation ID found!")

        row = -1
        for row, annot in enumerate(self.annotTableModel.annotationList):
            if annot.ID == annotation.ID:
                break
        assert(row > -1)

        self.annotListTblWdg.selectRow(row)
        self.mainTabs.setCurrentIndex(1)

    @pyqtSlot(object, object)
    def viewParameter(self, annotation, parameter):
        self.viewAnnotation(annotation)        
        self.modParamWgt.viewParameter(parameter)
        self.taggingTabs.setCurrentIndex(1)


    def refreshModelingParam(self):
        self.modParamWgt.loadModelingParameter()




    def editPreferences(self):
        settingsDlg = SettingsDlg(self.settings, self)
        if settingsDlg.exec_() == QDialog.Accepted:
            self.settings = getSettings()

            self.gitMng = GitManager(self.settings.config["GIT"])
            self.dbPath   = os.path.abspath(os.path.expanduser(self.settings.config["GIT"]["local"]))



    def setNeedSaving(self):
        if not self.needSavingDisabled:
            self.needSaving = True


    def annotTableLayoutChanged(self):
        self.selectedAnnotationChanged(self.annotListTblWdg.selectedIndexes())

    # Note: No @pyqtSlot because QModelIndexList doesn't exist in PyQt5.
    # @pyqtSlot(QModelIndexList)
    # @pyqtSlot(QItemSelection, QItemSelection)
    def selectedAnnotationChanged(self, selected, deselected=None):

        if not hasattr(self, "editAnnotWgt"):
            # This if is triggered when selectedAnnotationChanged is automatically called
            # in the construction of the dialog. We don't want to run this function in this 
            # context because not all the components of the dialog has been initialized yet.
            return

        ######## This block of code is to manage the cancelation of the change in selected
        ######## annotation when the saving of the previously selected annotation has 
        ######## been unsucessful and the user asked to correct the situation.
        if self.cancelledAnnotSelectinChange :
            self.cancelledAnnotSelectinChange = False
            return False

        if self.checkSavingAnnot() == False:
            if not deselected is None :
                self.cancelledAnnotSelectinChange = True
                self.annotListTblWdg.selectRow(deselected.indexes()[0].row())        
            return False
        ###################


        self.needSavingDisabled = True 
        self.editAnnotWgt.setDisabled(False)
        self.currentAnnotation = self.annotTableModel.getSelectedAnnotation(selected)
        if self.currentAnnotation is None:
            # The current index is invalid. Thus, we deactivate controls
            # used to modify the current annotation.
            self.clearAddAnnotation()
        else:
            self.editAnnotSubWgt.selectAnnotType(self.currentAnnotation.type)

            ## UPDATING EXPERIMENTAL PROPERTIES
            #self.expPropWgt.expPropertiesListModel.clear()
            #for prop in self.currentAnnotation.experimentProperties:
            #    self.expPropWgt.expPropertiesListModel.addProperty(prop.name, prop.value, prop.unit)
            self.expPropWgt.fillingExpPropList()   
            self.expPropWgt.expPropertiesListModel.refresh()


        self.refreshTagList()
        #self.tagAnnotGroupBox.setDisabled(self.currentAnnotation is None)    
        self.taggingTabs.setDisabled(self.currentAnnotation is None)    

        self.detectAnnotChange = False
        self.selectedAnnotationChangedConfirmed.emit()
        self.detectAnnotChange = True
            
        self.refreshModelingParam()
        self.needSavingDisabled = False 


    def builtOntoTrees(self, recompute=False):
        self.ontoMng = OntoManager(recompute=recompute)
        self.treeData                  = self.ontoMng.trees 
        self.dicData                   = self.ontoMng.dics
        #self.nlTreeModel               = TreeModel(self.treeData)
        #self.nlTreeView                = TreeView(self.nlTreeModel)
        #self.nlTreeView.clicked.connect(self.nlTreeWasClicked)


    #def nlTreeWasClicked(self, selected):
    #    tagId     = selected.data(Qt.UserRole)
    #    self.addTagToAnnotation(tagId)


    def getSelectedTags(self):
        return [self.selectedTagsWidget.itemWidget(self.selectedTagsWidget.item(i)).tag for i in range(self.selectedTagsWidget.count())]


    def getSuggestedTags(self):
        return [self.suggestedTagsWidget.itemWidget(self.suggestedTagsWidget.item(i)).tag for i in range(self.suggestedTagsWidget.count())]


    def tagSuggestionSelected(self, name):
        id = [tagId for tagId, tagName in self.dicData.items() if name == tagName]
        assert(len(id)==1)
        id = id[0]
        self.addTagToAnnotation(id)
        self.tagEdit.erase = True
        self.tagEdit.clearEditText()

    def editTextChanged(self, text):
        if self.tagEdit.erase:
            self.tagEdit.clearEditText()
            self.tagEdit.erase = False
            return True
        return False




    def addTagToSelected(self, tagId):
        # check that it is not a duplicate (for the widget)
        if not tagId in [tag.id for tag in self.getSelectedTags()]:
            tag = Tag(tagId, self.dicData[tagId])
            item = QListWidgetItem()
            tagEdit = TagWidget(tag, self.selectedTagsWidget)
            self.selectedTagsWidget.addItem(item)
            self.selectedTagsWidget.setItemWidget(item, tagEdit)
            tagEdit.clicked.connect(self.selectedTagClicked)

            # Check if the tag has been persisted for this paper    
            if self.IdTxt.text() in self.selectedTagPersist:
                tagEdit.persist = tagId in self.selectedTagPersist[self.IdTxt.text()]



    def addTagToAnnotation(self, tagId, tagName=None):
        if not tagId in self.currentAnnotation.tagIds:
            if tagName is None:
                tagName = self.dicData[tagId]                
            else:
                if not tagName in self.dicData:
                    self.dicData[tagId] = tagName
                    self.ontoMng.savePickle()
                    self.updateAutoCompleteTagList()

            self.currentAnnotation.addTag(tagId, tagName)
            self.needSaving = True

            self.tagSuggester.addUsedTag(tagId)
            self.refreshTagList()


    def addSuggestedTagFromId(self, tagId):
        tagName = self.dicData[tagId]
        tag = Tag(tagId, tagName)
        item = QListWidgetItem()
        tagEdit = TagWidget(tag, self.suggestedTagsWidget)
        self.suggestedTagsWidget.addItem(item)
        self.suggestedTagsWidget.setItemWidget(item, tagEdit)
        tagEdit.clicked.connect(self.suggestedTagClicked)
        tagEdit.persist = tagId in self.suggestTagPersist
    


    def selectedTagClicked(self, tag):
        modifiers = QApplication.keyboardModifiers()
        if modifiers == Qt.ShiftModifier:
            ID = self.IdTxt.text()
            if not ID in self.selectedTagPersist:
                self.selectedTagPersist[ID] = []
            if tag.id in self.selectedTagPersist[ID]:
                self.selectedTagPersist[ID].remove(tag.id)
            else:
                self.selectedTagPersist[ID].append(tag.id)
        
                # If there is already other annotations associated with this 
                # paper, ask if the persistence should also be applied to them.
                fileName = join(self.dbPath, Id2FileName(self.IdTxt.text())) + ".pcr"
                with open(fileName, "r", encoding="utf-8", errors='ignore') as f:
                    try:
                        annotations = Annotation.readIn(f)
                    except ValueError:
                        raise ValueError("Problem reading file " + fileName + ". The JSON coding of this file seems corrupted.")
            
                isNewAnnot = not self.currentAnnotation in self.annotTableModel.annotationList or self.currentAnnotation is None
                if len(annotations) > 1 - int(isNewAnnot) :
                    msgBox = QMessageBox(self)
                    msgBox.setWindowTitle("Tag persistence propagation")
                    msgBox.setText("There is already existing annotations for this paper. Do you want this tag to be added " +
                                   "to them too (any unsaved changes would be saved first)?")
                    msgBox.setStandardButtons(QMessageBox.No | QMessageBox.Yes)
                    msgBox.setDefaultButton(QMessageBox.No)
                    if msgBox.exec_() == QMessageBox.Yes:
                        # Add this tag to existing annotations (associated with the current publication)

                        # Save unsaved modifications if there are any...
                        if self.needSaving:
                            self.saveAnnotation()
                            with open(fileName, "r", encoding="utf-8", errors='ignore') as f:
                                try:
                                    annotations = Annotation.readIn(f)
                                except ValueError:
                                    raise ValueError("Problem reading file " + fileName + ". The JSON coding of this file seems corrupted.")

                        commit = False
                        for annot in annotations:
                            if not tag.id in [annotTag.id for annotTag in annot.tags]:
                                annot.addTag(tag.id, tag.name)
                                commit = True

                        if commit:
                            # TODO: Should be in a try block and if it generate an exception
                            # we should role back to last git version.
                            with open(fileName, "w", encoding="utf-8", errors='ignore') as f:
                                Annotation.dump(f, annotations)

                            self.gitMng.addFiles([fileName])
                            self.refreshListAnnotation()

            self.savePersistTag()
        else:
            self.removeTag(tag)

    def suggestedTagClicked(self, tag):
        modifiers = QApplication.keyboardModifiers()
        if modifiers == Qt.ShiftModifier:
            if tag.id in self.suggestTagPersist:
                self.suggestTagPersist.remove(tag.id)
            else:
                self.suggestTagPersist.append(tag.id)

            self.savePersistTag()
        else:
            self.removeSuggestedTag(tag)


    def removeTag(self, tag):
        for row, tagItem in enumerate(self.getSelectedTags()):
            if tagItem.id == tag.id:
                self.tagEdit.setFocus()
                self.selectedTagsWidget.takeItem(row)
                self.currentAnnotation.removeTag(tag.id)
                self.needSaving = True
                self.tagSuggester.removeUsedTag(tag.id)
        self.refreshTagList()



    def removeSuggestedTag(self, tag):
        self.addTagToAnnotation(tag.id)

    def openPDF(self):
        base_path = join(self.dbPath, Id2FileName(self.IdTxt.text()))
        pdf_path = base_path + ".pdf"

        if not os.path.isfile(pdf_path):
            msgBox = QMessageBox(self)
            msgBox.setWindowTitle("Missing PDF")
            msgBox.setText("The PDF file seems to be missing. Do you want to attach one?")
            msgBox.setStandardButtons(QMessageBox.No | QMessageBox.Yes)
            msgBox.setDefaultButton(QMessageBox.Yes)
            if msgBox.exec_() == QMessageBox.Yes:
                if not self.importPDF():
                    return
            else:
                return

        QDesktopServices.openUrl(QUrl.fromLocalFile(pdf_path))

    def checkSavingAnnot(self):
        if self.needSaving:
            msgBox = QMessageBox(self)
            msgBox.setWindowTitle("Unsaved annotation")
            msgBox.setText("There are unsaved changes in the currently loaded annotation. Do you want to save modifications?")
            msgBox.setStandardButtons(QMessageBox.No | QMessageBox.Yes)
            msgBox.setDefaultButton(QMessageBox.Yes)
            if msgBox.exec_() == QMessageBox.Yes:
                self.editAnnotSubWgt.commentEdt.setFocus()
                if self.saveAnnotation() == False:
                    return False
            self.needSaving = False
        return True

    @pyqtSlot(QItemSelection, QItemSelection)
    def paperSelectionChanged(self, selected, deselected):
        if selected.indexes():
            index = selected.indexes()[0]

            # FIXME DEBUG.
            # row = index.model().mapToSource(index).row()
            # print("\n")
            # print("TITLE: " + index.model().sourceModel()._zotero_wrap.reference_title(row))
            # print("PROXY ROW NB: " + str(index.row()))
            # print("SOURCE ROW NB: " + str(row))
            # /FIXME DEBUG.

            if self.checkSavingAnnot() == False:
                return False

            # FIXME Delayed refactoring. Simplify. Remove hard-coded column number.
            reference_id = index.model().data(index.model().index(index.row(), 0))

            if reference_id == "":
                msgBox = QMessageBox(self)
                msgBox.setWindowTitle("Warning")
                msgBox.setText("This paper has currently no ID. It will not be possible " +
                               "to process this paper until an ID is attributed. Would " +
                               "you like to set its ID now?")
                msgBox.setStandardButtons(QMessageBox.No | QMessageBox.Yes)
                msgBox.setDefaultButton(QMessageBox.Yes)
                if msgBox.exec_() == QMessageBox.Yes:
                    self.zotero_widget.edit_reference()
                    self.paperSelectionChanged(selected, deselected)
                else:
                    self.invalidPaperChoice()
                return

            isPMID = False
            isUNPUBLISHED = False
            isDOI = False
            if "PMID" in reference_id:
                isPMID = True
            elif "UNPUBLISHED" in reference_id:
                isUNPUBLISHED = True
            else:
                isDOI = True

            self.IdTxt.setText(reference_id)

            # Check if paper is already in the database
            if not self.checkIdInDB(self.IdTxt.text()):
                if isDOI or isPMID:
                    msgBox = QMessageBox(self)
                    msgBox.setWindowTitle("Paper not in the database")
                    msgBox.setText("This paper is not already in the curator database.")
                    pdfButton        = QPushButton("Select PDF")
                    msgBox.setStandardButtons(QMessageBox.Cancel)
                    msgBox.addButton(pdfButton, QMessageBox.YesRole)


                    if isDOI:
                        websiteButton    = QPushButton("Follow DOI to the publication website")
                        msgBox.addButton(websiteButton, QMessageBox.ActionRole)

                    msgBox.setDefaultButton(pdfButton)
                    retCode = msgBox.exec_()

                    if retCode == 0:
                        try:
                            if not self.importPDF():
                                self.invalidPaperChoice()
                                return
                        except UnicodeEncodeError:
                            errorMessage(self, "Unicode error", "Please check that " +\
                                         "the path of the file you are trying to " +\
                                         "upload does not contain non ASCII " +\
                                         "characters. Complete support of unicode " +\
                                         "encoding for file names and paths are " +\
                                         "not provided.")

                    elif retCode == 1 and isDOI:
                        url = "http://dx.doi.org/" + self.IdTxt.text()
                        webbrowser.open(url)
                        return
                    else:
                        self.invalidPaperChoice()
                        return

                elif isUNPUBLISHED:
                    saveFileName = join(self.dbPath, Id2FileName(self.IdTxt.text()))
                    with open(saveFileName + ".pcr", 'w', encoding="utf-8", errors='ignore'):
                        self.gitMng.addFiles([saveFileName + ".pcr"])


            self.openPDFBtn.setDisabled(isUNPUBLISHED)
            self.refreshListAnnotation(0)
            self.paperGroupBox.setDisabled(False)
            self.listAnnotGroupBox.setDisabled(False)


    def invalidPaperChoice(self):
        self.clearPaper()
        self.refreshListAnnotation()
        self.clearAddAnnotation()
        self.refreshTagList()
        self.editAnnotWgt.setDisabled(True)
        #self.tagAnnotGroupBox.setDisabled(True)    
        self.taggingTabs.setDisabled(True)    
        self.paperGroupBox.setDisabled(True)        
        self.listAnnotGroupBox.setDisabled(True)            


    def refreshTagList(self):
        self.refreshSelectedTagList()
        self.refreshSuggestedTagList()


    def refreshSelectedTagList(self):
        # Selected tag list
        self.selectedTagsWidget.clear()
        if not self.currentAnnotation is None:
            for id in self.currentAnnotation.tagIds:
                self.addTagToSelected(id)
        self.refreshModelingParam()    



    def refreshSuggestedTagList(self):
        # Suggested tag list
        self.suggestedTagsWidget.clear()
        if not self.currentAnnotation is None:
            annotationFileName = join(self.dbPath, Id2FileName(self.IdTxt.text())) + ".pcr"
            tagIds = self.tagSuggester.suggestions(annotationFileName, [tag.id for tag in self.getSelectedTags()], 200)

            unusedPersistedSuggestedTags = []
            selectedTags                 = [tag.id for tag in self.getSelectedTags()] 
            for persistId in self.suggestTagPersist:
                if persistId in tagIds:
                    tagIds.remove(persistId)
                if not persistId in selectedTags:
                    unusedPersistedSuggestedTags.append(persistId)



            for id in unusedPersistedSuggestedTags + tagIds:    
                self.addSuggestedTagFromId(id)


    def clearPaper(self):
        #self.refEdt.setText("")
        self.IdTxt.setText("")
        



    def waitForOCR(self, paperId, notify):
        
        while(not self.restClient.checkOCRFinished(paperId, self.dbPath)):
            self.window.statusLabel.setText("Performing OCR...")
            time.sleep(5)          
        
        self.window.statusLabel.setText("OCR finished.")
        if notify == QMessageBox.Yes:
            msgBox = QMessageBox()
            msgBox.setStandardButtons(QMessageBox.Cancel)
            msgBox.setWindowTitle("OCR process finished")
            msgBox.setText("The optical character recognition for the paper " + paperId
                           + " is finished. You can now start annotating this paper.")
            msgBox.exec_()


    def importPDF(self):
        # Import a PDF

        if not checkID(self.IdTxt.text()):
            errorMessage(self, "Error", "This ID seem to be invalid.")
            return

        fileName, _ = QFileDialog.getOpenFileName(self, 'Open file')
        fileName    = fileName.encode("utf-8").decode("utf-8")
        if fileName != '':
            saveFileName = join(self.dbPath, Id2FileName(self.IdTxt.text()))
            if os.path.isfile(saveFileName + ".txt"):
                errorMessage(self, "Error", "This PDF has already been imported to the database.")

            try:
                self.restClient.importPDF(fileName, self.IdTxt.text(), self.dbPath)
                
            except ConnectionError as e:
                errorMessage(self, "Error", "Failed to connect to the REST server. Error message: " + str(e))                                
                return False
                
            except RESTImportPDFErr as e:
                msgBox = QMessageBox(self)
                msgBox.setWindowTitle("This PDF needs OCR")
                msgBox.setText(str(e) + " Do you want to be notifed when this process is done?")
                msgBox.setStandardButtons(QMessageBox.No | QMessageBox.Yes)
                msgBox.setDefaultButton(QMessageBox.Yes)
                
                thread = Thread(target = self.waitForOCR, 
                                args = (self.IdTxt.text(), msgBox.exec_(), ))
                thread.start()  
                return False

            pcr_path = saveFileName + ".pcr"
            if not os.path.isfile(pcr_path):
                with open(pcr_path, "w", encoding="utf-8", errors="ignore"):
                    self.gitMng.addFiles([pcr_path])

            return True

        return False



    def pushToServer(self):
        info = self.gitMng.push()
        if info is None:
                        msgBox = QMessageBox(self)
                        msgBox.setWindowTitle("Push error")
                        msgBox.setText("The push operation has not been performed because you are in offline mode.")
                        msgBox.setStandardButtons(QMessageBox.Ok)
                        msgBox.exec_()

        elif info.flags & info.ERROR :
            msgBox = QMessageBox(self)
            msgBox.setWindowTitle("Push error")
            msgBox.setText("An error occured while trying to push to the server. Error flag: '" + str(info.flags) + "', message: '" + str(info.summary) + "'.")
            msgBox.setStandardButtons(QMessageBox.Ok)
            msgBox.exec_()
        else:
            msgBox = QMessageBox(self)
            msgBox.setWindowTitle("Repository pushed to the server")
            msgBox.setText("Modifications has been successfully pushed to the server.")
            msgBox.setStandardButtons(QMessageBox.Ok)
            msgBox.exec_()

        self.needPush = False



    def getCurrentContext(self):
        try:
            txtFileName = join(self.dbPath, Id2FileName(self.IdTxt.text())) + ".txt"
            with open(txtFileName, 'r', encoding="utf-8", errors='ignore') as f :
                fileText = f.read()
                contextStart = self.currentAnnotation.start - self.contextLength
                contextEnd = self.currentAnnotation.start + len(self.currentAnnotation.text) + self.contextLength
                return fileText[contextStart:contextEnd]
        except FileNotFoundError:
            return ""



    def deleteAnnotation(self):
        self.annotTableModel.annotationList.remove(self.currentAnnotation)
        self.currentAnnotation = None

        fileName = join(self.dbPath, Id2FileName(self.IdTxt.text())) + ".pcr"
        with open(fileName, "w", encoding="utf-8", errors='ignore') as f:
            Annotation.dump(f, self.annotTableModel.annotationList)

        self.clearAddAnnotation()
        self.gitMng.addFiles([fileName])
        self.needPush = True
        self.needSaving = False
        self.detectAnnotChange = False
        self.refreshListAnnotation()
        self.refreshModelingParam()



    def saveAnnotation(self):

        self.editAnnotSubWgt.updateCurrentAnnotation()
        fileName = join(self.dbPath, Id2FileName(self.IdTxt.text())) + ".pcr"

        with open(fileName, "r", encoding="utf-8", errors='ignore') as f:
            try:
                annots = Annotation.readIn(f)
            except ValueError:
                raise ValueError("Problem reading file " + fileName + ". The JSON coding of this file seems corrupted.")
            
        # Existing annotation has been modified
        if self.currentAnnotation is None:
            return
        if self.currentAnnotation.localizer is None:
            return


        self.currentAnnotation.experimentProperties = self.expPropWgt.getExpProperties()

        if self.currentAnnotation in self.annotTableModel.annotationList:
            for i, annot in enumerate(annots):
                if self.currentAnnotation.ID == annot.ID:
                    annots[i] = self.currentAnnotation
                    break
            row = None


        # New annotation has been created
        else:
            annots.append(self.currentAnnotation)

            # Select the new (last) annotation
            row = -1

        with open(fileName, "w", encoding="utf-8", errors='ignore') as f:
            Annotation.dump(f, annots)

        self.gitMng.addFiles([fileName])
        self.needPush = True
        self.detectAnnotChange = False
        self.needSaving = False
        self.refreshListAnnotation(row)
        return True

        

    def newAnnotation(self):
        if self.checkSavingAnnot() == False:
            return False

        self.clearAddAnnotation()
        self.annotListTblWdg.setCurrentIndex(QModelIndex())
        self.currentAnnotation = Annotation(pubId=self.IdTxt.text())
        if self.IdTxt.text() in self.selectedTagPersist:
            for id in self.selectedTagPersist[self.IdTxt.text()]:
                self.currentAnnotation.addTag(id, self.dicData[id])

        self.needSaving = False
        self.refreshTagList()
        self.taggingTabs.setEnabled(True)    
        self.refreshModelingParam()
        return True

        

    def duplicateAnnotation(self):
        if self.checkSavingAnnot() == False:
            return False

        #self.clearAddAnnotation()
        self.currentAnnotation = self.currentAnnotation.duplicate()
        #self.annotListTblWdg.setCurrentIndex(QModelIndex())
        self.saveAnnotation()
        #self.currentAnnotation = self.currentAnnotation.duplicate()
        #if self.IdTxt.text() in self.selectedTagPersist:
        #    for id in self.selectedTagPersist[self.IdTxt.text()]:
        #        self.currentAnnotation.addTag(id, self.dicData[id])

        #self.needSaving = False
        #self.refreshTagList()
        #self.taggingTabs.setEnabled(True)    
        #self.refreshModelingParam()
        return True

    def checkIdInDB(self, ID):
        # FIXME Delayed refactoring.
        papersPCR = glob(join(self.dbPath, Id2FileName(ID) + ".pcr")) # paper curation record
        papersPDF = glob(join(self.dbPath, Id2FileName(ID) + ".pdf")) # paper curation record
        papersTXT = glob(join(self.dbPath, Id2FileName(ID) + ".txt")) # paper curation record
        
        if len(papersPCR) == 0:
            return 0 # There is no record yet for this paper.
            
        if len(papersPDF) == 0 or len(papersTXT) == 0:
            return 1 # We have a record, but not pdf or txt.
            
        return 2 # We have a record, a txt and a pdf.


    def clearAddAnnotation(self):
        self.expPropWgt.fillingExpPropList(checkAll=True)
        self.currentAnnotation = None
        tmp = self.detectAnnotChange
        self.detectAnnotChange = False
        self.annotationCleared.emit()
        self.detectAnnotChange = tmp


    def refreshListAnnotation(self, row = None):

        self.annotTableModel.annotationList = []
        try :
            with open(join(self.dbPath, Id2FileName(self.IdTxt.text()) + ".pcr"), 'r', encoding="utf-8", errors='ignore') as f:        
                self.annotTableModel.annotationList = Annotation.readIn(f)    

            if not row is None:
                if row < 0:
                    self.annotListTblWdg.selectRow(self.annotListTblWdg.model().rowCount()-row)
                else:
                    self.annotListTblWdg.selectRow(row)

        except FileNotFoundError:
            # Set and invalid index
            self.annotListTblWdg.selectRow(-1)
        finally:
            self.selectedAnnotationChanged(self.annotListTblWdg.selectedIndexes())

            self.annotListTblWdg.sortByColumn(self.annotTableModel.sortCol, 
                                           self.annotTableModel.sortOrder)

            self.annotTableModel.refresh()







__author__ = "Pierre-Alexandre Fonta"

import os

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QFormLayout


def package_directory():
    """Return the absolute path to the directory containing the package files."""
    return os.path.abspath(os.path.dirname(__file__))


def configure_form_layout(form_layout):
    """Configure in an harmonized way QFormLayouts."""
    form_layout.setFormAlignment(Qt.AlignLeft | Qt.AlignTop)
    form_layout.setLabelAlignment(Qt.AlignLeft)
    form_layout.setRowWrapPolicy(QFormLayout.DontWrapRows)



#!/usr/bin/python3

__author__ = "Christian O'Reilly"

import configparser
import os

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (QLabel, QGridLayout, QGroupBox, QVBoxLayout,
                             QLineEdit, QCheckBox, QComboBox, QWidget,
                             QPushButton, QTabWidget, QDialog)
from neurocurator.utils import package_directory


def getSettings(popDialog = False):
    
    def popDialogFct():
        form = SettingsDlg()
        if form.exec_() == QDialog.Accepted:
            return getSettings()
        else:
            return None
    
    if popDialog:
        return popDialogFct()             
    
    try:
        return Settings()
    except FileNotFoundError:
        return popDialogFct()        
        

class Settings:
    fileName = os.path.join(package_directory(), 'settings.ini')
    def __init__(self):
        self.config = configparser.ConfigParser()
        with open(Settings.fileName) as configfile: # Raise an exception if file does not exist
            self.config.read(Settings.fileName)
            #self.config.read(configfile)

    def save(self):
        with open(Settings.fileName, 'w') as configfile:
          self.config.write(configfile)


class SettingsDlg(QDialog):

    restRoot = "https://bbpteam.epfl.ch"

    def __init__(self, settings=None, parent=None):
        super().__init__(parent)

        self.setWindowTitle("Curator settings")
        self.setGeometry(100, 300, 1000, 200)

        self.settings = settings

        self.projectSettings = ProjectSettings(self.settings)

        self.mainTabs = QTabWidget(self)
        self.mainTabs.addTab(self.projectSettings, "Projects")   

        if self.settings is None:
            self.restServerURLTxt = QLineEdit(SettingsDlg.restRoot + "/neurocurator/api/v1.0/", self)
        else:
            self.restServerURLTxt = QLineEdit(self.settings.config["REST"]["serverURL"], self)
            
        self.okBtn            = QPushButton('OK', self)

        # Signals
        self.okBtn.clicked.connect(self.writeConfig)

        # Layout
        layout = QVBoxLayout()

        # REST server
        self.restGroupBox = QGroupBox("REST server")
        grid = QGridLayout(self.restGroupBox)
        grid.addWidget(QLabel('REST server URL', self), 0, 0)
        grid.addWidget(self.restServerURLTxt, 0, 1)

        layout.addWidget(self.mainTabs)
        layout.addWidget(self.restGroupBox)
        layout.addWidget(self.okBtn)

        self.setLayout(layout)

        # Detect changes in git repository URL
        #self.gitProtocol.currentIndexChanged.connect(self.gitURLChanged)

    def writeConfig(self):
        config = configparser.ConfigParser()

        config['DEFAULT'] = {}

                
        config['REST'] = {'serverURL'      : self.restServerURLTxt.text()}

        if self.settings is None:
            config['WINDOW'] = {}
        elif "WINDOW" in self.settings.config: 
            config['WINDOW'] = self.settings.config["WINDOW"]
        else:
            config['WINDOW'] = {}

        config = self.projectSettings.writeConfig(config)

        with open(os.path.join(package_directory(), 'settings.ini'), 'w') as configfile:
          config.write(configfile)

        self.accept() 






class ProjectSettings(QWidget):

    def __init__(self, settings=None, parent=None):
        super().__init__(parent)

        self.zoteroLibraryTypeCB     = QComboBox(self)
        self.zoteroLibraryTypeCB.addItem("group")
        self.zoteroLibraryTypeCB.addItem("user")

        self.settings = settings

        # Creating widgets with initial values
        self.gitProtocol = QComboBox(self)
        protocols = ["http", "git+ssh"]
        self.gitProtocol.addItems(protocols)
        self.noRemotechkbox = QCheckBox("Don't use any remote", self)

        if self.settings is None:
            self.gitProtocol.setCurrentIndex(0)
            self.gitRemoteTxt     = QLineEdit('github.com/christian-oreilly/corpus-thalamus.git', self)
            self.gitLocalTxt      = QLineEdit(os.path.expanduser('~/curator_DB/'), self)
            self.gitUserTxt       = QLineEdit("git", self) #getpass.getuser(), self)

            self.zoteroLibIDTxt   = QLineEdit('427244', self)
            self.zoteroApiKeyTxt  = QLineEdit('4D3rDZsAVBd139alqoVZBKOO', self)
            self.zoteroLibraryTypeCB.setCurrentIndex(0)

            self.restServerURLTxt = QLineEdit(SettingsDlg.restRoot + "neurocurator/api/v1.0/", self)
            
            self.noRemotechkbox.setCheckState(Qt.Unchecked)
    
        else:
                
            if self.settings.config['GIT']['remote'] == "":
                self.gitRemoteTxt     = QLineEdit("", self)
                self.gitUserTxt       = QLineEdit("", self)
                self.noRemotechkbox.setCheckState(Qt.Checked)
          
            else:
                self.gitRemoteTxt     = QLineEdit(self.settings.config['GIT']['remote'], self)
                self.gitUserTxt       = QLineEdit(self.settings.config['GIT']['user'], self)
                self.noRemotechkbox.setCheckState(Qt.Unchecked)
                
                if "protocol" in self.settings.config['GIT']:
                    self.gitProtocol.setCurrentIndex(protocols.index(self.settings.config['GIT']['protocol']))
                else:
                    self.gitProtocol.setCurrentIndex(0)          
          
            self.gitLocalTxt      = QLineEdit(self.settings.config['GIT']['local'], self)

            self.zoteroLibIDTxt   = QLineEdit(self.settings.config['ZOTERO']['libraryID'], self)
            self.zoteroApiKeyTxt  = QLineEdit(self.settings.config['ZOTERO']['apiKey'], self)

            if self.settings.config['ZOTERO']['libraryType'] == "group":
                self.zoteroLibraryTypeCB.setCurrentIndex(0)
            elif self.settings.config['ZOTERO']['libraryType'] == "user":
                self.zoteroLibraryTypeCB.setCurrentIndex(1)
            else:
                raise ValueError

        self.zoteroLibraryIDInstructions = QLabel('', self)
        self.zoteroLibraryIDInstructions.setTextInteractionFlags(Qt.LinksAccessibleByMouse)
        self.zoteroLibraryIDInstructions.setOpenExternalLinks(True)

        # Signals
        self.zoteroLibraryTypeCB.currentIndexChanged.connect(self.updateZoteroLibraryIDInstructions)
        self.noRemotechkbox.stateChanged.connect(self.noRemoteChanged)

        # Layout
        layout = QVBoxLayout()

        # GIT
        self.gitGroupBox = QGroupBox("GIT")
        gridGIT = QGridLayout(self.gitGroupBox)
        
        gridGIT.addWidget(QLabel('Remote repository', self), 0, 0)
        gridGIT.addWidget(self.gitProtocol, 0, 1)
        gridGIT.addWidget(QLabel('://', self), 0, 2)
        gridGIT.addWidget(self.gitUserTxt, 0, 3)
        gridGIT.addWidget(QLabel('@', self), 0, 4)
        gridGIT.addWidget(self.gitRemoteTxt, 0, 5)
        
        gridGIT.addWidget(self.noRemotechkbox, 1, 0, 1, 6)
                
        gridGIT.addWidget(QLabel('Local repository', self), 2, 0)
        gridGIT.addWidget(self.gitLocalTxt, 2, 1, 1, 5)

        # Zotero
        self.zoteroGroupBox = QGroupBox("Zotero")
        gridGIT = QGridLayout(self.zoteroGroupBox)
        gridGIT.addWidget(QLabel('Library type', self), 0, 0)
        gridGIT.addWidget(self.zoteroLibraryTypeCB, 0, 1)
        gridGIT.addWidget(QLabel('Library ID', self), 1, 0)
        gridGIT.addWidget(self.zoteroLibIDTxt, 1, 1)
        gridGIT.addWidget(self.zoteroLibraryIDInstructions, 2, 1, 1, 2)
        gridGIT.addWidget(QLabel('API Key', self), 3, 0)
        gridGIT.addWidget(self.zoteroApiKeyTxt, 3, 1)
        privateKeyInst = QLabel("You're private key can be generate here: <a href=\"https://www.zotero.org/settings/keys/new\">https://www.zotero.org/settings/keys/new</a>", self)
        privateKeyInst.setTextInteractionFlags(Qt.LinksAccessibleByMouse)
        privateKeyInst.setOpenExternalLinks(True)
        gridGIT.addWidget(privateKeyInst, 4, 1, 1, 2)

        self.updateZoteroLibraryIDInstructions()

        layout.addWidget(self.gitGroupBox)
        layout.addWidget(self.zoteroGroupBox)

        self.setLayout(layout)

        # Detect changes in git repository URL
        #self.gitProtocol.currentIndexChanged.connect(self.gitURLChanged)

    def noRemoteChanged(self, checked):
        self.gitProtocol.setDisabled(self.noRemotechkbox.checkState())
        self.gitUserTxt.setDisabled(self.noRemotechkbox.checkState())
        self.gitRemoteTxt.setDisabled(self.noRemotechkbox.checkState())



    def updateZoteroLibraryIDInstructions(self):
        if self.zoteroLibraryTypeCB.currentText() == "group":
            self.zoteroLibraryIDInstructions.setText('The ID can be found by opening the group’s page (links to your groups page should be ' +
                                                     '<a href=\"https://www.zotero.org/groups/\">here</a>), and hovering\n' +
                                                     'over the group settings link. The ID is the integer after ' +
                                                     '/groups/. You must be the owner of the group to see the setting link.')
        elif self.zoteroLibraryTypeCB.currentText() == "user":
            self.zoteroLibraryIDInstructions.setText('Personal libary ID can be found here: ' + 
                                                     '<a href=\"https://www.zotero.org/settings/keys\">https://www.zotero.org/settings/keys</a>')
        else:
            raise ValueError


    def writeConfig(self, config):

        config['GIT'] = {'protocol'        : self.gitProtocol.currentText(),
                         'remote'          : self.gitRemoteTxt.text(),
                         'local'           : self.gitLocalTxt.text(),
                         'user'            : self.gitUserTxt.text()}

        config['ZOTERO'] = {'libraryID'    : self.zoteroLibIDTxt.text(),
                            'apiKey'       : self.zoteroApiKeyTxt.text(),
                            'libraryType'  : self.zoteroLibraryTypeCB.currentText()}
                
        return config



# -*- coding: utf-8 -*-
"""
Created on Tue Jul 26 10:23:39 2016

@author: oreilly
"""

from sys import platform as _platform

from PyQt5.QtCore import (QModelIndex, pyqtSignal, pyqtSlot, QAbstractTableModel,
                          QEvent, Qt)
from PyQt5.QtWidgets import (QGridLayout, QAbstractItemView, QTableView, QWidget,
                             QLineEdit)

from nat import ontoServ


class OntoAutoComplete(QLineEdit):

    completionTerminated = pyqtSignal(dict)

    def focusInEvent(self, event):
        if _platform == "linux" or _platform == "linux2":
            # This behavior is the original one, which is fine in linux. On MacOS,
            # selecting an item change the value of the combobox and then set the focus on it
            # which was making this line erase the selected text.
            self.setText("")

        super().focusInEvent(event)

        
    def event(self, event):
        if event.type() == QEvent.KeyPress:
            if event.key() == Qt.Key_Tab:
                completion = ontoServ.autocomplete(self.text())
                self.completionTerminated.emit(completion)
                
        return super().event(event)


class OntoOnlineSearch(QWidget):

    tagSelected = pyqtSignal(str, str)
    
    def __init__(self, parent=None):
        super().__init__(parent)

        self.termListTblWdg = QTableView()
        self.termTableModel = OntoTermsListModel(parent=self)
        self.termListTblWdg.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.termListTblWdg.setSelectionMode(QAbstractItemView.SingleSelection)
        self.termListTblWdg.setModel(self.termTableModel)

        self.termListTblWdg.setColumnWidth(0, 300)
        self.termListTblWdg.setColumnWidth(1, 200)

        # Signals
        self.termSelectionModel = self.termListTblWdg.selectionModel()
        self.termSelectionModel.selectionChanged.connect(self.termSelected)
        #self.termTableModel.layoutChanged.connect(self.termTableLayoutChanged)

        self.autoCompleteTxt = OntoAutoComplete(self)
        self.autoCompleteTxt.completionTerminated.connect(self.completionUpdate)
        
        grid = QGridLayout(self)
        grid.addWidget(self.termListTblWdg, 0, 0)   
        grid.addWidget(self.autoCompleteTxt, 1, 0)

    def termSelected(self, selection, deselected=None):
        #id = [tagId for tagId, tagName in self.dicData.items() if name == tagName]
        #assert(len(id)==1)
        #id = id[0]
        #self.addTagToAnnotation(id)
        #self.tagEdit.erase = True
        #self.tagEdit.clearEditText()
        if len(selection.indexes()):
            term = self.termTableModel.getTerm(selection.indexes()[0])
            self.tagSelected.emit(term[0], term[1])
    
            # Unselect selected item in the table view.
            #self.termListTblWdg.setCurrentCell(-1,-1)
            self.termListTblWdg.clearSelection()
            self.termTableModel.refresh()

    @pyqtSlot(dict)
    def completionUpdate(self, termDic):
        #print(termDic)
        self.termTableModel.setTerms(termDic)


class OntoTermsListModel(QAbstractTableModel):

    def __init__(self, terms = [], header = ['term', 'curie'], parent=None):
        super().__init__(parent)
        self.terms = terms
        self.header = header
        self.sortCol   = 0 
        self.sortOrder = Qt.AscendingOrder

    def rowCount(self, parent=QModelIndex()):
        return len(self.terms)

    def columnCount(self, parent=QModelIndex()):
        return len(self.header)


    """
    def getSelectedTerm(self, selection):

        if isinstance(selection, list):
            if selection == []:
                return None
            elif isinstance(selection[0], QModelIndex):
                index = selection[0]
        else:
            if selection.at(0) is None:
                return None
            index = selection.at(0).indexes()[0]
        return self.annotationList[index.row()]
    """


    def getByIndex(self, term, ind):
        return term[ind]


    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None

        if role != Qt.DisplayRole:
            return None

        return self.getByIndex(self.terms[index.row()], index.column())

    def headerData(self, col, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.header[col]
        return None

    def sort(self, col=None, order=None):
        if col is None:
            col = self.sortCol 
        if order is None:
            order = self.sortOrder

        # Sort table by given column number col.
        self.layoutAboutToBeChanged.emit()
        reverse = (order == Qt.DescendingOrder)
        self.annotationList = sorted(self.terms, key=lambda x: self.getByIndex(x, col), reverse = reverse) #operator.itemgetter(col))
        #if order == Qt.DescendingOrder:
        #    self.mylist.reverse()
        self.layoutChanged.emit()

    def getTerm(self, index):
        return self.terms[index.row()]

    def setTerms(self, termDic):
        self.terms = [(term, curie) for curie, term in termDic.items()]
        self.refresh()
    
    def refresh(self):
        self.layoutChanged.emit()



import operator
import pickle
from copy import deepcopy

import numpy as np

from nat.annotation import Annotation


class TagSuggester:

    def __init__(self):
        self.usedTag = {}
        self.globalVsLocalRatio = 0.5

    def addUsedTag(self, tagId):
        if tagId in self.usedTag:
            self.usedTag[tagId] += 1
        else:
            self.usedTag[tagId] = 1
        self.save()


    def removeUsedTag(self, tagId):
        if tagId in self.usedTag:
            self.usedTag[tagId] -= 1
            self.save()



    def suggestions(self, annotationFileName, selectedIds, numberOfSuggestions=30):
        if len(self.usedTag) == 0:
            return []

        tagScores = deepcopy(self.usedTag)
        
        # Computing global indice
        maxGlobalUse = np.max(list(tagScores.values())) / self.globalVsLocalRatio

        # To avoid zero division
        if maxGlobalUse == 0:
            maxGlobalUse = 1.0 / self.globalVsLocalRatio

        for key in tagScores:
            tagScores[key] /= maxGlobalUse 



        # Computing local indices
        localScores = {}
        try :
            with open(annotationFileName, 'r', encoding="utf-8", errors='ignore') as f:        
                annots = Annotation.readIn(f)
                for annot in annots:
                    for tag in annot.tags:
                        if tag.id in localScores:
                            localScores[tag.id] += 1
                        else:
                            localScores[tag.id] = 1
        except FileNotFoundError:
            pass
        
        if len(localScores):
            maxLocallUse = np.max(list(localScores.values())) / (1.0 - self.globalVsLocalRatio)
            for key in localScores:
                if not key in tagScores:
                    tagScores[key]  = localScores[key] / maxLocallUse
                else:
                    tagScores[key] += localScores[key] / maxLocallUse


        for key in selectedIds:
            if key in tagScores:
                del tagScores[key]


        if len(tagScores):
            keys, values = zip(*sorted(tagScores.items(), key=operator.itemgetter(1) , reverse=True))
            return list(keys[:min(numberOfSuggestions, len(keys))])
        else:
            return []


    @staticmethod
    def load(fileName="suggester.pickle"):
        try:
            with open(fileName, 'rb') as f:
                return pickle.load(f)
        except:
            return TagSuggester()

    def save(self, fileName="suggester.pickle"):
        try:
            with open(fileName, 'wb') as f:
                pickle.dump(self, f)
        except:
            pass



    



#!/usr/bin/python3

__author__ = 'oreilly'
__email__  = 'christian.oreilly@epfl.ch'

from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtGui import QColor, QPalette
from PyQt5.QtWidgets import QApplication, QLabel

from nat.tag import Tag


class TagWidget(QLabel):
    
    clicked = pyqtSignal(Tag)
    
    def __init__(self, tag, parent=None):
        super().__init__(tag.name, parent)
        self.tag = tag
        self.setAutoFillBackground(True)
        self.persist = False

    def mouseReleaseEvent(self, event):
        modifiers = QApplication.keyboardModifiers()
        if modifiers == Qt.ShiftModifier:
            self.persist = not self.persist
        self.clicked.emit(self.tag)

    @property
    def persist(self):
        return self.__persist

    @persist.setter
    def persist(self, shouldPersist):
        if shouldPersist:
            ligthRed = QColor(255, 153, 153)
            palette = QPalette()
            palette.setColor(self.backgroundRole(),ligthRed)
            self.setPalette(palette)        
        else:
            #lightGreen = QtGui.QColor(191, 237, 135)
            palette = QPalette()
            palette.setColor(self.backgroundRole(), QColor(255, 255, 255))
            self.setPalette(palette)
        self.__persist = shouldPersist





#!/usr/bin/python3

__author__ = "Christian O'Reilly"

from PyQt5.QtCore import Qt, QAbstractTableModel
from PyQt5.QtCore import pyqtSignal, pyqtSlot, QModelIndex
from PyQt5.QtWidgets import QMessageBox, QTableView

from nat.modelingParameter import (getParameterTypes, getParameterTypeIDFromName,
                                   getParameterTypeFromID, ParameterTypeTree)
from nat.paramDesc import ParamDescFunction, ParamDescTrace
from nat.values import ValuesSimple, ValuesCompound
from nat.variable import Variable, NumericalVariable
from .itemDelegates import (DoubleDelegate, ParamTypeDelegate, UnitDelegate,
                            StatisticsDelegate)


parameterTypeTree = ParameterTypeTree.load()
parameterTypes = getParameterTypes()


class VariableTableView(QTableView):

    depTypeSelected  = pyqtSignal(str)
    #selectionChanged = QtCore.Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setItemDelegate(DoubleDelegate(self))
        typeDelegate = ParamTypeDelegate(self)
        self.setItemDelegateForRow(0, typeDelegate)
        self.setItemDelegateForRow(1, UnitDelegate(self))
        self.setItemDelegateForRow(2, StatisticsDelegate(self))
        typeDelegate.typeSelected.connect(self.typeSelected)

    @pyqtSlot(str)
    def typeSelected(self, paramType):
        if self.selectionModel().currentIndex().column() == 0:
            self.depTypeSelected.emit(paramType)

    def deleteVariable(self):
        self.model().deleteVariable(self.selectionModel().currentIndex().column())

    def deleteSample(self):
        self.model().deleteSample(self.selectionModel().currentIndex().row())


class VariableListModel(QAbstractTableModel):

    def __init__(self, colHeader = ['Dependant', 'Independant 1'], rowHeader = ['Type', 'Unit', 'Statistic'], parent=None):
        super().__init__(parent)
        self.clear(colHeader, rowHeader)

    def clear(self, colHeader = ['Dependant', 'Independant 1'], rowHeader = ['Type', 'Unit', 'Statistic']):
        self.colHeader = colHeader
        self.rowHeader = rowHeader
        self.nbSample  = 0
        self.nbDep     = 1
        self.nbIndep   = 1
        self.__data       = {} 
        for col in colHeader:        
            for row in rowHeader:
                if row == "Statistic":
                    self.__data[(row, col)] = "raw"
                elif row == "Unit":
                    self.__data[(row, col)] = "dimensionless"        
                else:
                    self.__data[(row, col)] = None
        self.refresh()



    def setFromParam(self, param):
        if isinstance(param.description, ParamDescFunction):
            self.__data = {}
            self.__data[('Type',       'Dependant')] = getParameterTypeFromID(param.description.depVar.typeId).name
            self.__data[('Unit',       'Dependant')] = param.description.depVar.unit
            self.__data[('Statistic',  'Dependant')] = param.description.depVar.statistic
            self.colHeader = ['Dependant']

            for colInd in range(1, len(param.description.indepVars)+1):
                colName = 'Independant ' + str(colInd)        
                self.colHeader.append(colName)
                self.__data[('Type',       colName)] = getParameterTypeFromID(param.description.indepVars[colInd-1].typeId).name
                self.__data[('Unit',       colName)] = param.description.indepVars[colInd-1].unit
                self.__data[('Statistic', colName)] = param.description.indepVars[colInd-1].statistic

        elif isinstance(param.description, ParamDescTrace):
            self.__data = {}      
            valuesObject = param.description.depVar.values
            self.__data[('Type', 'Dependant')] = getParameterTypeFromID(param.description.depVar.typeId).name    
            
            #### Add the appropriate number of samples
            self.nbSample = 0
            self.rowHeader = self.rowHeader[:3]
            for noSample in range(len(param.description.indepVars[0].values.values)):
                self.addSample(refresh=False)            
            
            self.colHeader = ['Dependant']
            
            #### ValuesCompound as dependant variable
            if isinstance(valuesObject, ValuesCompound):
                self.nbDep = len(valuesObject.valueLst)
                for ind, val in enumerate(valuesObject.valueLst):
                    if ind == 0:                       
                        varName = "Dependant"                    
                    else:
                        varName = "Dependant comp. " + str(ind+1)   
                        self.colHeader.append(varName)
                        self.__data[('Type', varName)] = self.__data[('Type', 'Dependant')]  
                        
                    self.__data[('Unit',       varName)] = val.unit
                    self.__data[('Statistic',  varName)] = val.statistic

                    for noSample, sample in enumerate(val.values):
                        self.__data[(str(noSample+1), varName)] = sample

            #### ValuesSimple as dependant variable  
            elif isinstance(valuesObject, ValuesSimple):

                self.__data[('Unit',       'Dependant')] = valuesObject.unit
                self.__data[('Statistic',  'Dependant')] = valuesObject.statistic
    
                self.nbDep = 1
                for noSample, sample in enumerate(valuesObject.values):
                    self.__data[(str(noSample+1), 'Dependant')] = sample

            #### Independant variables

            self.nbIndep = len(param.description.indepVars)
            for colInd in range(self.nbDep, self.nbIndep+self.nbDep):
                colName = 'Independant ' + str(colInd)        
                self.colHeader.append(colName)
                self.__data[('Type',       colName)] = getParameterTypeFromID(param.description.indepVars[colInd-self.nbDep].typeId).name
                self.__data[('Unit',       colName)] = param.description.indepVars[colInd-self.nbDep].values.unit
                self.__data[('Statistic', colName)] = param.description.indepVars[colInd-self.nbDep].values.statistic

                for noSample, sample in enumerate(param.description.indepVars[colInd-self.nbDep].values.values):
                    self.__data[(str(noSample+1), colName)] = sample
        else:
            raise TypeError

        self.refresh()

    def getIndepVars(self, varType="Variable"):
        if varType == "Variable":
            return [Variable(getParameterTypeIDFromName(self.__data[('Type',      varLabel)]), 
                                                        self.__data[('Unit',      varLabel)], 
                                                        self.__data[('Statistic', varLabel)])
                        for varLabel in self.colHeader if "Independant" in varLabel]

        elif varType == "NumericalVariable":
            indepVars = []
            for varLabel in self.colHeader:
                if not "Independant" in varLabel:
                    continue
                floatValues = []
                for row in self.rowHeader[3:]:
                    try:
                        floatValues.append(float(self.__data[(row, varLabel)]))
                    except ValueError:
                        floatValues.append(float('nan'))

                values = ValuesSimple(floatValues, self.__data[('Unit',      varLabel)], 
                                                    self.__data[('Statistic', varLabel)])

                typeId = getParameterTypeIDFromName(self.__data[('Type', varLabel)])
                indepVars.append(NumericalVariable(typeId, values))
            return indepVars

        else:
            raise ValueError



    def getDepVar(self, varType="Variable"):
        
        if varType == "Variable":
            typeId = getParameterTypeIDFromName(self.__data[('Type', 'Dependant')])
            return Variable(typeId, 
                            self.__data[('Unit',      'Dependant')], 
                            self.__data[('Statistic', 'Dependant')])
                            
        elif varType == "NumericalVariable":
            depVarValues = []
            
            for varLabel in self.colHeader:
                if not "Dependant" in varLabel:
                    continue            
                if varLabel == "Dependant" :
                    typeId  = getParameterTypeIDFromName(self.__data[('Type', varLabel)])
                
                floatValues = []
                for row in self.rowHeader[3:]:
                    try:
                        floatValues.append(float(self.__data[(row, varLabel)]))
                    except ValueError:
                        floatValues.append(float('nan'))
    
                values = ValuesSimple(floatValues, self.__data[('Unit',      varLabel)], 
                                                   self.__data[('Statistic', varLabel)])
    
                if self.nbDep == 1 :
                    return NumericalVariable(typeId, values)   
                else:
                    depVarValues.append(values)
            
            values = ValuesCompound(depVarValues)
            
            return NumericalVariable(typeId, values)   
                
        else:
            raise ValueError



    def rowCount(self, parent=QModelIndex()):
        return len(self.rowHeader)

    def columnCount(self, parent=QModelIndex()):
        return len(self.colHeader)


    def getType(self, col):
        return self.__data["Type", self.colHeader[col]]

    def getUnit(self, col):
        return self.__data["Unit", self.colHeader[col]]

    def getStatistic(self, col):
        return self.__data["Statistic", self.colHeader[col]]


    def addVariable(self):      
        inc = 1
        col = 'Independant ' + str(self.nbIndep+inc)
        while col in self.colHeader:
            inc += 1
            col = 'Independant ' + str(self.nbIndep+inc)
            
        self.colHeader.append(col)
        for row in self.rowHeader:
            if row == "Statistic":
                self.__data[(row, col)] = "raw"
            elif row == "Unit":
                self.__data[(row, col)] = "dimensionless"                
            else:
                self.__data[(row, col)] = None

        self.nbIndep += 1  
        self.refresh()


    def addDepCompnent(self):
        inc = 1
        col = 'Dependant comp. ' + str(self.nbDep+inc)
        while col in self.colHeader:
            inc += 1
            col = 'Dependant comp. ' + str(self.nbDep+inc)
                    
        self.colHeader.insert(self.nbDep, col)
        for row in self.rowHeader:
            if row == "Statistic":
                self.__data[(row, col)] = "raw"
            elif row == "Unit":
                self.__data[(row, col)] = "dimensionless"     
            elif row == "Type":
                self.__data[(row, col)] = self.__data[(row, 'Dependant')]                
            else:
                self.__data[(row, col)] = None

        self.nbDep += 1
        self.refresh()




    def addSample(self, refresh=True):
        self.nbSample += 1
        row = str(self.nbSample)
        self.rowHeader.append(row)
        for col in self.colHeader:
            self.__data[(row, col)] = None

        if refresh:
            self.refresh()



    def deleteVariable(self, col):

        msgBox = QMessageBox()
        msgBox.setWindowTitle("Deletion")
        msgBox.setText("Are you sure you want to delete the variable '" + self.colHeader[col] + "'?")
        msgBox.setStandardButtons(QMessageBox.No | QMessageBox.Yes)
        msgBox.setDefaultButton(QMessageBox.No)
        if msgBox.exec_() == QMessageBox.Yes:
            for rowName in self.rowHeader:
                del self.__data[(rowName, self.colHeader[col])]
                
            if "Independant" in self.colHeader[col]:
                self.nbIndep -= 1
            else:
                self.nbDep -= 1 
                
            del self.colHeader[col]
        self.refresh()
        
        

    def deleteSample(self, noRow):

        msgBox = QMessageBox()
        msgBox.setWindowTitle("Deletion")
        msgBox.setText("Are you sure you want to delete the sample at index " + self.rowHeader[noRow] + "?")
        msgBox.setStandardButtons(QMessageBox.No | QMessageBox.Yes)
        msgBox.setDefaultButton(QMessageBox.No)
        if msgBox.exec_() == QMessageBox.Yes:
            for colName in self.colHeader:
                del self.__data[(self.rowHeader[noRow], colName)]
            del self.rowHeader[noRow]

        self.refresh()


    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None

        if role != Qt.DisplayRole and  role != Qt.EditRole:
            return None

        return str(self.__data[(self.rowHeader[index.row()], self.colHeader[index.column()])])



    def setData(self, index, value, role=Qt.DisplayRole):
        if value is None:
            value = ""
            
        if self.colHeader[index.column()] == "Dependant" and self.rowHeader[index.row()] == "Type":
            for colName in self.rowHeader:
                if 'Dependant comp. ' in colName:
                    self.__data[(self.rowHeader[index.row()], colName)] = value
                        
        self.__data[(self.rowHeader[index.row()], self.colHeader[index.column()])] = value
        return True


    def flags(self, index):
        if "comp." in self.colHeader[index.column()] and self.rowHeader[index.row()] == "Type":
            return Qt.NoItemFlags
            
        else:
            return Qt.ItemIsEditable | Qt.ItemIsEnabled


    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.colHeader[section]        
        elif orientation == Qt.Vertical and role == Qt.DisplayRole:
            return self.rowHeader[section]
        return None


    def refresh(self):
        self.layoutChanged.emit()

    def load(self, param):
        #param.description.
        ### TODO: LOAD FROM PARAM
        self.refresh()




# -*- coding: utf-8 -*-
"""
Created on Mon Jun 20 16:23:49 2016

@author: oreilly
"""

from PyQt5.QtCore import Qt, QModelIndex, QItemSelection, QAbstractTableModel


class AnnotationListModel(QAbstractTableModel):

    def __init__(self, annotationList = [], header = ['ID', 'type', 'localizer', 'comment'], parent=None):
        super().__init__(parent)
        self.annotationList = annotationList
        self.header = header
        self.nbCol = len(header)
        self.sortCol   = 0 
        self.sortOrder = Qt.AscendingOrder

    def rowCount(self, parent=QModelIndex()):
        return len(self.annotationList)

    def columnCount(self, parent=QModelIndex()):
        return self.nbCol 

    def getSelectedAnnotation(self, selected):
        # FIXME Delayed refactoring. Different entry points with different types.
        if isinstance(selected, QItemSelection):
            selected = selected.indexes()

        # At this stage, selected is a QModelIndexList in both cases.
        if selected:
            index = selected[0]
            try:
                return self.annotationList[index.row()]
            except IndexError:
                # FIXME Delayed refactoring. The selection has not been properly
                # cleared (annotation deletion).
                return None
        else:
            return None

    def getByIndex(self, annot, ind):
        if ind == 0:
            return annot.ID
        elif ind == 1:
            return str(annot.type)
        elif ind == 2:
            return str(annot.localizer)
        elif ind == 3 :
            return annot.comment
        else:
            raise ValueError

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None

        if role != Qt.DisplayRole:
            return None

        return self.getByIndex(self.annotationList[index.row()], index.column())

    def headerData(self, col, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.header[col]
        return None

    def sort(self, col=None, order=None):
        if col is None:
            col = self.sortCol 
        if order is None:
            order = self.sortOrder

        # Sort table by given column number col.
        self.layoutAboutToBeChanged.emit()

        reverse = (order == Qt.DescendingOrder)
        self.annotationList = sorted(self.annotationList, key=lambda x: self.getByIndex(x, col), reverse = reverse) #operator.itemgetter(col))
        #if order == Qt.DescendingOrder:
        #    self.mylist.reverse()

        self.layoutChanged.emit()

    def refresh(self):
        self.layoutChanged.emit()



#!/usr/bin/python3

__author__ = "Christian O'Reilly"

from PyQt5.QtWidgets import QHBoxLayout, QAbstractItemView, QWidget

from nat.annotation import getParametersForPub
from nat.paramDesc import ParamRef
from nat.utils import Id2FileName

from .paramFunctionWgt import ParameterInstanceTableView, ParameterInstanceListModel


class ExpPropWgt(QWidget):

    def __init__(self, parent=None):
        super().__init__(parent)

        self.main_window = parent

        # Widgets
        self.expPropertiesListTblWdg = ParameterInstanceTableView()
        self.expPropertiesListModel = ParameterInstanceListModel(parent=self)
        self.expPropertiesListTblWdg.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.expPropertiesListTblWdg.setSelectionMode(QAbstractItemView.SingleSelection)
        self.expPropertiesListTblWdg.setModel(self.expPropertiesListModel)

        self.expPropertiesListTblWdg.setColumnWidth(0, 20)
        self.expPropertiesListTblWdg.setColumnWidth(1, 150)
        self.expPropertiesListTblWdg.setColumnWidth(2, 500)

        # Signal
        self.expPropertiesListTblWdg.tableCheckBoxClicked.connect(self.propSelectionChanged)

        # Layout        
        expPropertiesLayout = QHBoxLayout(self)
        expPropertiesLayout.addWidget(self.expPropertiesListTblWdg)

    def fillingExpPropList(self, checkAll=False):
        if self.main_window.currentAnnotation is None:
            return
            
        parameters = getParametersForPub(self.main_window.dbPath, Id2FileName(self.main_window.currentAnnotation.pubId))
        parameters = [param for param in parameters if param.isExperimentProperty == True]

        if checkAll:
            selectedParams = [param.id for param in parameters]  
        elif self.main_window.currentAnnotation is None:
            selectedParams = []      
        else:
            selectedParams = [ref.instanceId for ref in self.main_window.currentAnnotation.experimentProperties]
            
        self.expPropertiesListModel.load(parameters, selectedParams)
        self.expPropertiesListModel.refresh()
        

    def getExpProperties(self):
        parameters = []
        for noRow in range(self.expPropertiesListModel.rowCount()):
            paramInstanceId = self.expPropertiesListModel.getParamInstanceID(noRow)
            if self.expPropertiesListModel.selected[paramInstanceId]:
                parameters.append(ParamRef(paramInstanceId, 
                                       self.expPropertiesListModel.getParamTypeID(noRow)))
        return parameters

    def propSelectionChanged(self, row):
        self.main_window.setNeedSaving()



#!/usr/bin/env python3.4

__authors__ = ["Christian O'Reilly", "Pierre-Alexandre Fonta"]
__maintainer__ = "Pierre-Alexandre Fonta"

import sys

from PyQt5.QtWidgets import QApplication

from neurocurator.mainWin import Window


def main():
    app = QApplication(sys.argv)
    window = Window()
    # FIXME DEBUG.
    # window.setGeometry(0, 0, 500, 1424)
    # /FIXME DEBUG.
    window.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



__authors__ = ["Pierre-Alexandre Fonta", "Christian O'Reilly"]
__maintainer__ = "Pierre-Alexandre Fonta"

from PyQt5.QtCore import QModelIndex, Qt, QAbstractTableModel
from PyQt5.QtGui import QColor, QBrush

from nat.annotationSearch import AnnotationSearch


class ZoteroTableModel(QAbstractTableModel):

    HEADERS = ["ID", "Title", "Creators", "Year", "Journal", "Annotations"]

    def __init__(self, zotero_wrap, check_id_fct, annotations_path, parent=None):
        super().__init__(parent)
        # FIXME Delayed refactoring.
        self.check_id_fct = check_id_fct
        # FIXME Delayed refactoring.
        self.annotations_path = annotations_path
        self._zotero_wrap = zotero_wrap
        # TODO Cache them?
        self._annotation_counts = []
        # TODO For performance, create a data structure with only the displayed data?

    # Data I/O methods section.

    def load(self):
        """Load the Zotero data and compute the annotation counts."""
        self.layoutAboutToBeChanged.emit()
        # TODO Implement an offline mode. Catch PyZoteroError.
        self._zotero_wrap.initialize()
        self._compute_annotation_counts()
        self.layoutChanged.emit()

    def refresh(self):
        """Replace the cached Zotero data with the distant one."""
        # TODO Displayed annotation counts will be inconsistent if new ones
        # have been created during the refresh!
        # TODO Only latest sorting kept. Row numbers change after in the proxy
        # model in this case. Enforce all previous sorting after refresh()?
        self.layoutAboutToBeChanged.emit()
        # TODO Implement an offline mode. Catch PyZoteroError.
        self._zotero_wrap.load_distant()
        self._compute_annotation_counts()
        self.layoutChanged.emit()

    # Qt interface implementation section.

    def columnCount(self, parent=QModelIndex()):
        if parent.isValid():
            return 0
        else:
            return len(self.HEADERS)

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None
        row = index.row()
        column = index.column()
        if self._is_index_too_large(row, column):
            return None
        if role == Qt.DisplayRole:
            return self._cell_data(row, column)
        if role == Qt.BackgroundRole:
            ref_id = self._zotero_wrap.reference_id(row)
            if self.check_id_fct(ref_id) == 2:
                color = QColor(191, 237, 135)
                return QBrush(color, Qt.SolidPattern)
            elif self.check_id_fct(ref_id) == 1:
                color = QColor(150, 150, 150)
                return QBrush(color, Qt.SolidPattern)
            else:
                return None
        return None

    def flags(self, index):
        if not index.isValid():
            return Qt.ItemIsEnabled
        elif self.HEADERS[index.column()] == "Annotations":
            return super().flags(index) | Qt.ItemIsEditable
        else:
            return super().flags(index)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if section >= len(self.HEADERS):
            return None
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.HEADERS[section]
        return None

    def rowCount(self, parent=QModelIndex()):
        if parent.isValid():
            return 0
        else:
            return self._zotero_wrap.reference_count()

    def setData(self, index, value, role=Qt.EditRole):
        if not index.isValid():
            return False
        row = index.row()
        column = index.column()
        if self._is_index_too_large(row, column):
            return False
        if role == Qt.EditRole:
            if self.HEADERS[column] == 'Annotations':
                self._set_annotation_count(row, value)
                self.dataChanged.emit(index, index)
                return True
        return False

    # Public methods section.

    def add_reference(self, ref):
        """Add the reference at the end. Return the row as a QModelIndex."""
        new_row = self.rowCount()
        self.beginInsertRows(QModelIndex(), new_row, new_row)
        self._zotero_wrap.create_local_reference(ref)
        self._insert_annotation_count()
        self.endInsertRows()
        # NB: Column number will not be used.
        return self.index(new_row, 0, QModelIndex())

    def update_reference(self, row, ref):
        """Update the reference at the given row."""
        self._zotero_wrap.update_local_reference(row, ref)
        start_index = self.index(row, 0, QModelIndex())
        end_index = self.index(row, self.columnCount() - 2, QModelIndex())
        self.dataChanged.emit(start_index, end_index)

    # Private @properties surrogates section.

    def _annotation_count(self, row):
        """Return the number of annotations of the reference at the given row."""
        return self._annotation_counts[row]

    def _set_annotation_count(self, row, count):
        """Set the number of annotations of the reference at the given row."""
        self._annotation_counts[row] = count

    def _insert_annotation_count(self):
        """Insert an annotation count of value 0 at the end, for a new reference."""
        self._annotation_counts.append(0)

    # Private methods section.

    def _compute_annotation_counts(self):
        """Compute the number of annotations for all references and set it."""
        # FIXME Delayed refactoring (related to search).
        results = AnnotationSearch(self.annotations_path).search()
        counts = results["Publication ID"].value_counts().to_dict()
        self._annotation_counts = [int(counts.get(self._zotero_wrap.reference_id(i), 0))
                                   for i in range(self._zotero_wrap.reference_count())]

    def _is_index_too_large(self, row, column):
        """Check if row and column numbers are not out of range.

        QModelIndex::isValid() only checks if the index belongs to a model and
        has non-negative row and column numbers.
        """
        return row >= self.rowCount() or column >= len(self.HEADERS)

    def _cell_data(self, row, column):
        """Return the attribute, in this column, of the reference in this row."""
        header = self.HEADERS[column]
        if header == "ID":
            return self._zotero_wrap.reference_id(row)
        elif header == "Title":
            return self._zotero_wrap.reference_title(row)
        elif header == "Creators":
            return self._zotero_wrap.reference_creator_surnames_str(row)
        elif header == "Year":
            # Return int for sorting.
            return self._zotero_wrap.reference_year(row)
        elif header == "Journal":
            return self._zotero_wrap.reference_journal(row)
        elif header == 'Annotations':
            # Return int for sorting.
            return self._annotation_count(row)



#!/usr/bin/python3

__author__ = "Christian O'Reilly"

import numpy as np
import pandas as pd
from PyQt5.QtCore import QModelIndex, pyqtSignal, pyqtSlot, Qt, QAbstractTableModel
from PyQt5.QtWidgets import (QTableView, QCheckBox, QVBoxLayout, QWidget,
                             QHBoxLayout, QGroupBox, QComboBox, QLineEdit,
                             QFileDialog, QSplitter, QPushButton,
                             QAbstractItemView, QHeaderView)

from nat.annotationSearch import (parameterKeys, annotationKeys,
                                  ParameterSearch,
                                  AnnotationSearch, parameterResultFields,
                                  annotationResultFields)
from nat.condition import ConditionAtom, ConditionAND, ConditionOR, ConditionNOT
from nat.ontoManager import OntoManager
from .autocomplete import AutoCompleteEdit
from .itemDelegates import ParamTypeCbo, CheckBoxDelegate


class SearchWgt(QWidget):

    annotationSelected = pyqtSignal(object)
    parameterSelected  = pyqtSignal(object, object)

    def __init__(self, searchType="Parameter", parent=None):
        super().__init__(parent)

        # FIXME Delayed refactoring.
        self._parent         = parent

        self.searchType     = searchType
        self.queryDef       = QueryDefinitionWgt(searchType, self)
        self.outputFormat   = OutputFormatWgt(searchType, self)
        
        self.view = QTableView()
        self.view.setWordWrap(True)
        self.view.setTextElideMode(Qt.ElideMiddle)
        self.view.resizeRowsToContents()
        self.view.verticalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        
        self.model = PandasModel()
        self.view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.view.setSelectionMode(QAbstractItemView.SingleSelection)

        self.view.setModel(self.model)

        self.buttonWgt  = QWidget(self)
        buttonLayout    = QHBoxLayout(self.buttonWgt)
        self.searchBtn  = QPushButton("Search", self)
        self.saveBtn    = QPushButton("Save as .csv", self)
        buttonLayout.addWidget(self.searchBtn)
        buttonLayout.addWidget(self.saveBtn)
        
        self.searchBtn.clicked.connect(self.search)  
        self.saveBtn.clicked.connect(self.saveResults)   
        self.view.doubleClicked.connect(self.loadItem)

        self.splitter = QSplitter(Qt.Vertical, self)
        
        self.splitter.addWidget(self.queryDef)
        self.splitter.addWidget(self.outputFormat)
        self.splitter.addWidget(self.view)
        self.splitter.addWidget(self.buttonWgt)    
        
        layout = QVBoxLayout(self)
        layout.addWidget(self.splitter)

    def loadItem(self, index):
        if self.searchType == "Parameter":
            parameter  = self.model.getObject(index, "obj_parameter")
            annotation = self.model.getObject(index, "obj_annotation")
            self.parameterSelected.emit(annotation, parameter)

        elif self.searchType == "Annotation":
            annotation = self.model.getObject(index, "obj_annotation")
            self.annotationSelected.emit(annotation)
        else:
            raise ValueError

        
    def search(self):        

        if self._parent is None:
            dbPath = None
        elif not hasattr(self._parent, "dbPath"):
            dbPath = None
        else:
            dbPath = self._parent.dbPath

        if self.searchType == "Parameter":
            searcher = ParameterSearch(dbPath)
        elif self.searchType == "Annotation":
            searcher = AnnotationSearch(dbPath)
        else:
            raise ValueError

        query = self.queryDef.getQuery()
        if not query is None:
            searcher.setSearchConditions(query)
        searcher.setResultFields(self.outputFormat.getFields())
        self.outputFormat.setSearcherProperties(searcher)
            
        result = searcher.search()
        self.model._data = result
        self.model.refresh()
        
    def saveResults(self):
        
        fname, _ = QFileDialog.getSaveFileName(self, 'Save research results')
        if not fname == "":
            if not "." in fname:
                fname = fname + ".csv"
            self.model._data.to_csv(fname)


class QueryDefinitionWgt(QGroupBox):

    def __init__(self, searchType, parent=None):
        super().__init__("Search conditions", parent)
        self.searchType = searchType
        self.node = QueryNodeWgt(self.searchType, self)       
        layout = QVBoxLayout(self)
        layout.addWidget(self.node)
        layout.addStretch(100)
        
    def getQuery(self):
        return self.node.getQuery()
    

class QueryNodeWgt(QWidget):

    valueTypeChanged = pyqtSignal(object, object)

    def __init__(self, searchType, parent=None):
        super().__init__(parent)
        self.conditionTypeCbo = QComboBox(self)
        self.rowsWgt          = QWidget(self)
        self.conditionTypeCbo.addItems(["", "AND", "OR", "NOT"])
        self.searchType = searchType
        self.queryRows  = []
        layout = QHBoxLayout(self)
        layout.addWidget(self.conditionTypeCbo)
        layout.addWidget(self.rowsWgt)
        self.rowsLayout = QVBoxLayout(self.rowsWgt)
        self.addARow()
        self.conditionTypeCbo.currentIndexChanged.connect(self.conditionTypeChanged)

    def getQuery(self):
        query = None
        if self.conditionTypeCbo.currentText() == "":
            if not (self.rowsLayout.count() == 1 and isinstance(self.queryRows[0], QueryRowWgt)):
                raise ValueError
            query = self.queryRows[0].getQuery()

        elif self.conditionTypeCbo.currentText() == "NOT":
            if not (self.rowsLayout.count() == 1 and isinstance(self.queryRows[0], QueryRowWgt)):
                raise ValueError
            query = self.queryRows[0].getQuery()
            if not query is None:
                query = ConditionNOT(query)

        elif self.conditionTypeCbo.currentText() == "AND": 
            query = [row.getQuery() for row in self.queryRows if not row.getQuery() is None]
            if len(query):
                query = ConditionAND(query)
                    
        elif self.conditionTypeCbo.currentText() == "OR": 
            query = [row.getQuery() for row in self.queryRows if not row.getQuery() is None]
            if len(query):
                query = ConditionOR(query)
            
        else:
            raise ValueError        
        
        return query
        
    def conditionTypeChanged(self):
        if self.conditionTypeCbo.currentText() in ["", "NOT"]:
            if not (self.rowsLayout.count() == 1 and isinstance(self.queryRows[0], QueryRowWgt)):
                self.clear()
                self.addARow()
                
        elif self.conditionTypeCbo.currentText() in ["AND", "OR"]: 
            if isinstance(self.queryRows[0], QueryRowWgt):
                self.clear()
                self.addANode()            
        else:
            raise ValueError


    def clear(self):
        while self.rowsLayout.count():
            child = self.rowsLayout.takeAt(0)
            child.widget().deleteLater()
        self.queryRows = []

        
    def addARow(self):
        rowWgt = QueryRowWgt(self.searchType, self)
        self.queryRows.append(rowWgt)
        self.rowsLayout.addWidget(rowWgt)   
        rowWgt.valueTypeChanged.connect(self.valueTypeChangedSlot)

    def addANode(self):
        nodeWgt = QueryNodeWgt(self.searchType, self)
        self.queryRows.append(nodeWgt)
        self.rowsLayout.addWidget(self.queryRows[-1])
        nodeWgt.valueTypeChanged.connect(self.valueTypeChangedSlot)

    @pyqtSlot(object)
    @pyqtSlot(object, object)
    def valueTypeChangedSlot(self, rowObject, nodeObject=None):
        if self.conditionTypeCbo.currentText() in ["", "NOT"]:
            self.valueTypeChanged.emit(rowObject, self)
        else:
            noRow = self.queryRows.index(nodeObject)
            if rowObject.valueType.currentText() == "" and self.rowsLayout.count() > 1 :
                child = self.rowsLayout.takeAt(noRow)
                child.widget().deleteLater()    
                del self.queryRows[noRow]
            elif rowObject.valueType.currentText() != "" and noRow == self.rowsLayout.count()-1 :
                #if self.conditionTypeCbo.currentText() in ["AND", "OR"]:
                self.addANode()


class QueryRowWgt(QWidget):

    valueTypeChanged = pyqtSignal(object)
    
    ontoMng  = OntoManager()
    treeData = ontoMng.trees 
    dicData  = ontoMng.dics

    tagNames  = np.array(list(dicData.values()))      
    del dicData
    del treeData    
    
    def __init__(self, searchType, parent=None):
        super().__init__(parent)
        
        self.searchType = searchType
        self.valueType  = QComboBox(self)
        self.valueType.addItem("")

        if self.searchType == "Annotation":
            self.valueType.addItems(annotationKeys)    
        elif self.searchType == "Parameter":
            self.valueType.addItems(parameterKeys)    
        else:
            raise ValueError()
            
        self.valueType.currentIndexChanged.connect(self.valueTypeChangedEmit)
        
        self.value      = QLineEdit(self)
        self.value.setEnabled(False)

        self.layout = QHBoxLayout(self)
        self.layout.addWidget(self.valueType)
        self.layout.addWidget(self.value)

    def valueTypeChangedEmit(self):
        self.value.setEnabled(self.valueType.currentText() != "")            

        child = self.layout.takeAt(1)
        child.widget().deleteLater()    
        if self.valueType.currentText() == "Has parameter":
            self.value = QComboBox(self)
            self.value.addItems(["False", "True"])
        elif self.valueType.currentText() == "Annotation type":
            self.value = QComboBox(self)
            self.value.addItems(["text", "figure", "table", "equation", "position"])
        elif self.valueType.currentText() == "Parameter name":
            self.value = ParamTypeCbo(self)
        elif self.valueType.currentText() in ["Tag name", "Required tag name"]:
            self.value = AutoCompleteEdit(self)
            self.value.setModel(QueryRowWgt.tagNames)                        
        elif self.valueType.currentText() == "Result type":
            self.value = QComboBox(self)
            self.value.addItems(["pointValue", "function", "numericalTrace"])
        else:
            self.value = QLineEdit(self)
 
        self.layout.addWidget(self.value)
        self.valueTypeChanged.emit(self)
        


    def getQuery(self):
        if self.valueType.currentText() == "":
            return None
            
        if isinstance(self.value, QLineEdit):
            value = self.value.text()
        elif isinstance(self.value, QComboBox):
            value = self.value.currentText()
        else:
            raise TypeError
            
        return ConditionAtom(self.valueType.currentText(), value)


class OutputFormatWgt(QGroupBox):

    def __init__(self, searchType, parent=None):
        super().__init__("Output format", parent)

        self.searchType = searchType

        if self.searchType == "Parameter":
            self.fields = parameterResultFields
        elif self.searchType == "Annotation":
            self.fields = annotationResultFields
        else:
            raise ValueError

        self.fieldsTblWdg       = FieldTableView(self)
        self.paramListModel     = FieldListModel(parent=self)
        #self.paramListTblWdg.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
        #self.paramListTblWdg.setSelectionMode(QtGui.QAbstractItemView.SingleSelection)
        self.fieldsTblWdg.setModel(self.paramListModel)
        self.fieldsTblWdg.setColumnWidth(0, 30)
        self.fieldsTblWdg.setColumnWidth(1, 200)

        self.outputProperties = OutputPropertiesWgt(searchType, self)

        layout = QHBoxLayout(self)
        layout.addWidget(self.fieldsTblWdg)
        layout.addWidget(self.outputProperties)
        
        self.paramListModel.load(self.fields)
    
    def getFields(self):
        return self.paramListModel.getSelectedFields()

    def setSearcherProperties(self, searcher):
        self.outputProperties.setSearcherProperties(searcher)


class OutputPropertiesWgt(QWidget):

    def __init__(self, searchType, parent=None):
        super().__init__(parent)

        self.searchType = searchType

        layout = QVBoxLayout(self)

        if self.searchType == "Parameter":
            self.expandRequiredTagsChk  = QCheckBox("Expand required tags")
            self.onlyCentralTendancyChk = QCheckBox("Show only central tendency of parameter values")
            layout.addWidget(self.onlyCentralTendancyChk)
            layout.addWidget(self.expandRequiredTagsChk)
            layout.addStretch(1)
            
        elif self.searchType == "Annotation":
            pass
        
        else:
            raise ValueError

    def setSearcherProperties(self, searcher):

        if self.searchType == "Parameter":
            searcher.expandRequiredTags  = self.expandRequiredTagsChk.isChecked()
            searcher.onlyCentralTendancy = self.onlyCentralTendancyChk.isChecked()
            
        elif self.searchType == "Annotation":
            pass
        
        else:
            raise ValueError


class PandasModel(QAbstractTableModel):

    def __init__(self, data=pd.DataFrame(), parent=None):
        super().__init__(parent)
        self._data = data

    def rowCount(self, parent=QModelIndex()):
        return len(self._data.values)

    def columns(self):
        return [col for col in self._data.columns if col[:4] != "obj_"] 
        
    def indDisplayColumns(self):
        return [no for no, col in enumerate(self._data.columns) if col[:4] != "obj_"] 

    def columnCount(self, parent=QModelIndex()):
        return len(self.columns())

    def data(self, index, role= Qt.DisplayRole):
        if index.isValid():
            if role == Qt.DisplayRole:
                return str(self._data.values[index.row()][self.indDisplayColumns()[index.column()]])
        return None

    def getObject(self, index, objField):
        for col in self._data.columns:
            if col == objField: 
                return self._data[objField][index.row()]
        raise ValueError

    def headerData(self, index, orientation, role=Qt.DisplayRole):
        if   orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.columns()[index]
        elif orientation == Qt.Vertical and role == Qt.DisplayRole:
            return str(self._data.index[index])
        return None


    def refresh(self):
        self.layoutChanged.emit()


class FieldTableView(QTableView):

    tableCheckBoxClicked = pyqtSignal(int)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setItemDelegateForColumn(0, CheckBoxDelegate(self))      

    @pyqtSlot(bool)
    def checkBoxClicked(self, clicked):
        self.model().toggleParameter(self.sender().row)
        self.tableCheckBoxClicked.emit(self.sender().row)


class FieldListModel(QAbstractTableModel):

    def __init__(self, colHeader = ['', 'Fields to include'], parent=None): #'Type',
        super().__init__(parent)

        #self.parameterList = parameterList
        self.colHeader = colHeader
        self.nbCol     = len(colHeader)
        #self.types       = []
        self.fields    = []
        self.selected  = []

    #def clear(self):
    #    for id in self.selected:
    #        self.selected[id] = False
    #    self.refresh()

    def rowCount(self, parent=QModelIndex()):
        return len(self.fields)

    def columnCount(self, parent=QModelIndex()):
        return self.nbCol 


    def getByIndex(self, row, col):
        if col == 0:
            return self.selected[row]
        elif col == 1:
            return self.fields[row]
        else:
            raise ValueError


    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None

        if role != Qt.DisplayRole:
            return None

        return self.getByIndex(index.row(), index.column())


    def flags(self, index):
        return Qt.ItemIsEditable | Qt.ItemIsEnabled



    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.colHeader[section]        
        return None


    def load(self, fields):
        self.fields = fields
        self.selected = [True]*len(fields)
        self.refresh()

        
    def selectParameter(self, row, selected):
        self.selected[row] = selected

    def toggleParameter(self, row):
        self.selected[row] = not self.selected[row]

    def refresh(self):
        self.layoutChanged.emit()

    def getSelectedFields(self):
        return [field for field, selected in zip(self.fields, self.selected) if selected]




#!/usr/bin/python3

__author__ = "Christian O'Reilly"

from PyQt5.QtCore import pyqtSignal, pyqtSlot, QItemSelection, QModelIndex
from PyQt5.QtWidgets import QWidget, QAbstractItemView, QPushButton, QGridLayout

from nat.modelingParameter import getParameterTypes
from nat.paramDesc import ParamDescTrace
from nat.parameterInstance import ParameterInstance

from .variableTableWgt import VariableTableView, VariableListModel


class ParamTraceWgt(QWidget):

    paramTypeSelected = pyqtSignal(str)
    
    def __init__(self, parent=None):
        super().__init__(parent)

        self.parameterTypes = getParameterTypes()

        # Widgets        
        self.varListTblWdg = VariableTableView(self)
        self.varListModel = VariableListModel(parent=self)
        self.varListTblWdg.setSelectionMode(QAbstractItemView.SingleSelection)
        self.varListTblWdg.setSelectionBehavior(QAbstractItemView.SelectItems)
        self.varListTblWdg.setModel(self.varListModel)

        selection = self.varListTblWdg.selectionModel()
        selection.selectionChanged.connect(self.varSelectionChanged)
        
        self.varListTblWdg.depTypeSelected.connect(self.newParamSelected)
        
        self.addIndepVarBtn    = QPushButton("Add indep. variable")
        self.addDepVarCompBtn  = QPushButton("Add a component to the dep. var.")
        self.deleteIndepVarBtn = QPushButton("Delete variable")
        self.addSample         = QPushButton("Add sample")
        self.deleteSample      = QPushButton("Delete sample")
        self.loadCSV           = QPushButton("Load CSV")

        self.deleteSample.setDisabled(True)
        self.deleteIndepVarBtn.setDisabled(True)
        
        # Signals
        self.addIndepVarBtn.clicked.connect(self.varListModel.addVariable)
        self.addDepVarCompBtn.clicked.connect(self.varListModel.addDepCompnent)
        self.deleteIndepVarBtn.clicked.connect(self.varListTblWdg.deleteVariable)
        self.addSample.clicked.connect(self.varListModel.addSample)
        self.deleteSample.clicked.connect(self.varListTblWdg.deleteSample)
        self.varListTblWdg.depTypeSelected.connect(self.depVarSelected)

        # Layout
        grid = QGridLayout(self)
        grid.addWidget(self.varListTblWdg, 0, 0, 7, 1)
        grid.addWidget(self.addIndepVarBtn, 0, 1)
        grid.addWidget(self.addDepVarCompBtn, 1, 1)
        grid.addWidget(self.deleteIndepVarBtn, 2, 1)
        grid.addWidget(self.addSample, 3, 1)
        grid.addWidget(self.deleteSample, 4, 1)
        grid.addWidget(self.loadCSV, 5, 1)

        self.varListTblWdg.clicked.connect(self.tableClicked)

    def tableClicked(self, index):
        self.varSelectionChanged(index, None)


    def newParamSelected(self, paramType):
        self.paramTypeSelected.emit(paramType)

    def varSelectionChanged(self, selected, deselected):
        #col = self.varListTblWdg.selectionModel().currentIndex().column()

        if isinstance(selected, QModelIndex):
            index = selected
        elif isinstance(selected, QItemSelection):
            if len(selected.indexes()) == 1:
                index = selected.indexes()[0]
            else:
                return
            
        self.deleteSample.setDisabled(index.row() < 3)
        colName = self.varListModel.colHeader[index.column()]
        nbIndep = self.varListModel.nbIndep
        self.deleteIndepVarBtn.setDisabled(colName == "Dependant" or 
                                           ("Independant" in colName and nbIndep == 1))

        #paramType = self.varListModel.getType(col)
        #if not paramType is None and "Dependant" in colName:
        #    self.paramTypeSelected.emit(paramType)


    @pyqtSlot(str)
    def depVarSelected(self, depVar):
        self.paramTypeSelected.emit(depVar)

    def newParameter(self):
        self.varListModel.clear()

    def saveParameter(self, relationship, paramId):
        depVar        = self.varListModel.getDepVar(varType="NumericalVariable")
        indepVars     = self.varListModel.getIndepVars(varType="NumericalVariable")
        description     = ParamDescTrace(depVar, indepVars)
        return ParameterInstance(paramId, description, [], relationship)


    def loadRow(self, currentParameter):
        if currentParameter is None:
            self.varListModel.clear()            
        else:
            self.varListModel.setFromParam(currentParameter)
            self.varListModel.refresh()

    def loadModelingParameter(self, row = None):
        pass





#!/usr/bin/python3

__author__ = "Christian O'Reilly"

from PyQt5.QtCore import QModelIndex, pyqtSignal, pyqtSlot, QAbstractTableModel, Qt
from PyQt5.QtWidgets import (QMessageBox, QGridLayout, QAbstractItemView,
                             QGroupBox, QSplitter, QVBoxLayout, QLabel,
                             QHBoxLayout, QStackedWidget, QCheckBox, QComboBox,
                             QTableView, QWidget, QPushButton)

from nat.modelingParameter import (getParameterTypeFromName,
                                   getParameterTypeNameFromID,
                                   getParameterTypeFromID)
from nat.ontoManager import OntoManager
from nat.tag import RequiredTag
from nat.tagUtilities import nlx2ks
from .itemDelegates import ReqTagDelegate
from .paramFunctionWgt import ParamFunctionWgt
from .paramRelationWgt import ParamRelationWgt
from .paramTraceWgt import ParamTraceWgt
from .paramValueWgt import ParamValueWgt


class RequiredTagsTableView(QTableView):

    setReqTags = pyqtSignal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.reqTagDelegate = ReqTagDelegate(self)
        self.setItemDelegateForColumn(1, self.reqTagDelegate)
        ontoMng = OntoManager()
        self.treeData                  = ontoMng.trees 
        self.dicData                   = ontoMng.dics
        self.reqTagDelegate.cboNeedPopulation.connect(self.setReqTags)

    def setReqTags(self, tagName):

        try:
            tagId = list(self.dicData.keys())[list(self.dicData.values()).index(tagName)]
        except ValueError:
            ontoMng       = OntoManager(recompute=True)      
            self.treeData = ontoMng.trees 
            self.dicData  = ontoMng.dics    
            tagId = list(self.dicData.keys())[list(self.dicData.values()).index(tagName)]
            
        
        if not tagId in self.treeData:
            raise ValueError("The term id " + tagId + " was not specified as an ontological root.")

        self.reqTagDelegate.addItems(list(self.treeData[tagId].values()))




class RequiredTagsListModel(QAbstractTableModel):

    def __init__(self, colHeader = ['Required categories', 'Selected tag'], parent=None):
        super().__init__(parent)
        self.colHeader             = colHeader
        self.nbCol                 = len(colHeader)
        self.requiredTagsIds       = []
        self.requiredTagsNames     = []
        self.selectedTagsIds       = []
        self.selectedTagsNames     = []
        ontoMng = OntoManager()
        self.treeData                  = ontoMng.trees 
        self.dicData                   = ontoMng.dics


    def rowCount(self, parent=QModelIndex()):
        return len(self.requiredTagsNames)

    def columnCount(self, parent=QModelIndex()):
        return self.nbCol 

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None

        if not role in [Qt.DisplayRole, Qt.UserRole]:
            return None

        if index.column() == 0:
            return self.requiredTagsNames[index.row()]
        elif index.column() == 1:
            return self.selectedTagsNames[index.row()]
        else:
            return None


    def setData(self, index, value, role=Qt.DisplayRole):
        if value is None:
            value = ""

        if index.column() == 0:
            self.requiredTagsNames[index.row()] = value
            tagId = list(self.dicData.keys())[list(self.dicData.values()).index(value)]            
            self.requiredTagsIds[index.row()]   = tagId
            
        elif index.column() == 1:
            if self.checkTagValidity(index.row(), value):
                self.selectedTagsNames[index.row()] = value
                tagId = list(self.dicData.keys())[list(self.dicData.values()).index(value)]    
                self.selectedTagsIds[index.row()]   = tagId

    def checkTagValidity(self, row, tagName):

        tagId = self.requiredTagsIds[row]

        if not tagId in self.treeData:    
            ontoMng       = OntoManager(recompute=True)      
            self.treeData = ontoMng.trees 
            self.dicData  = ontoMng.dics            
            if not tagId in self.treeData:                
                raise ValueError("Tag '" + tagId + "' is not a treeData root. TreeData roots are the following:" + str(list(self.treeData.keys())))

        return tagName in list(self.treeData[tagId].values())


    def flags(self, index):
        if index.column() == 0:
            result = super().flags(index)
            result &= ~Qt.ItemIsEditable
            return result
        else:
            return Qt.ItemIsEditable | Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.colHeader[section]        
        return None

    def refresh(self):
        self.layoutChanged.emit()

    def addTag(self, requiredTagId, requiredTagName, selectedTagId, selectedTagName):
        self.requiredTagsNames.append(requiredTagName)
        self.requiredTagsIds.append(requiredTagId)
        self.selectedTagsNames.append(selectedTagName)
        self.selectedTagsIds.append(selectedTagId)
        self.refresh()

    def clear(self):
        self.requiredTagsNames  = []
        self.requiredTagsIds    = []
        self.selectedTagsNames  = []
        self.selectedTagsIds    = []
        self.refresh()

    def getRequiredTags(self):
        return [RequiredTag(id, name, rootId) for id, name, rootId in zip(self.selectedTagsIds, self.selectedTagsNames, self.requiredTagsIds)]





class ParamModWgt(QWidget):

    def __init__(self, parent=None):
        super().__init__(parent)

        self.main_window = parent

        self.buildRequiredTagsGB()

        # Widgets        
        self.newParamBtn        = QPushButton("New")
        self.deleteParamBtn     = QPushButton("Delete")
        self.paramSaveAnnotBtn  = QPushButton("Save")
        buttonWidget             = QWidget(self)

        self.existingParamsGB     = QGroupBox("Existing parameters", self)

        self.paramListTblWdg      = QTableView()
        self.paramListModel     = ParameterListModel(parent=self)
        self.paramListTblWdg.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.paramListTblWdg.setSelectionMode(QAbstractItemView.SingleSelection)
        self.paramListTblWdg.setModel(self.paramListModel)

        self.paramListTblWdg.setColumnWidth(0, 150)
        self.paramListTblWdg.setColumnWidth(1, 350)

        self.relationWgt    = ParamRelationWgt(parent)
        self.newParamsGB    = QGroupBox("Parameter details", self)
        self.resultTypeCbo  = QComboBox(self)
        self.isExpProp      = QCheckBox("is an experimental property", self)

        self.resultTypeCbo.addItems(["point value", "function", "numerical trace"])
        
        self.singleValueParamWgt = ParamValueWgt(parent)
        self.functionParamWgt    = ParamFunctionWgt(parent)
        self.traceParamWgt       = ParamTraceWgt(parent)

        self.functionParamWgt.mainWgt = self

        self.paramModStack       = QStackedWidget(self)
        self.paramModStack.addWidget(self.singleValueParamWgt)
        self.paramModStack.addWidget(self.functionParamWgt)
        self.paramModStack.addWidget(self.traceParamWgt)

        # Signals
        selectionModel = self.paramListTblWdg.selectionModel()
        selectionModel.selectionChanged.connect(self.selectedParameterChanged)

        self.newParamBtn.clicked.connect(self.newParameter)
        self.deleteParamBtn.clicked.connect(self.deleteParameter)
        self.paramSaveAnnotBtn.clicked.connect(self.saveParameter)
        self.resultTypeCbo.currentIndexChanged.connect(self.paramModStack.setCurrentIndex)
        self.singleValueParamWgt.paramTypeSelected.connect(self.newParamTypeSelected)
        self.functionParamWgt.paramTypeSelected.connect(self.newParamTypeSelected)
        self.traceParamWgt.paramTypeSelected.connect(self.newParamTypeSelected)

        # Layout
        buttonLayout = QVBoxLayout(buttonWidget)
        buttonLayout.addWidget(self.paramSaveAnnotBtn)
        buttonLayout.addWidget(self.deleteParamBtn)
        buttonLayout.addWidget(self.newParamBtn)

        existGrid     = QHBoxLayout(self.existingParamsGB)
        existGrid.addWidget(buttonWidget)
        existGrid.addWidget(self.paramListTblWdg)
        
        newGrid     = QGridLayout(self.newParamsGB)
        newGrid.addWidget(QLabel("Result type"), 0, 0)
        newGrid.addWidget(self.resultTypeCbo, 0, 1)
        newGrid.addWidget(self.isExpProp, 0, 2)
        
        newGrid.addWidget(self.paramModStack, 1, 0, 1, 3)
        newGrid.addWidget(self.relationWgt, 1, 3)

        layout = QVBoxLayout(self)
        self.rootLayout = QSplitter(Qt.Vertical, self)
        self.rootLayout.setOrientation(Qt.Vertical)
        self.rootLayout.addWidget(self.existingParamsGB)
        self.rootLayout.addWidget(self.newParamsGB)
        self.rootLayout.addWidget(self.requireTagGB)
        layout.addWidget(self.rootLayout)

        # Initial behavior
        self.newParamBtn.setEnabled(True)
        self.deleteParamBtn.setEnabled(False)
        self.paramSaveAnnotBtn.setEnabled(False)
        self.additionMode = False
        self.newParamsGB.setEnabled(False)

    def setRootLayoutSizes(self, sizes):
        self.rootLayout.setSizes(sizes)


    def viewParameter(self, parameter):
        row = -1
        for row, param in enumerate(self.paramListModel.parameterList):
            if param.id == parameter.id:
                break
        assert(row > -1)
        self.paramListTblWdg.selectRow(row)

        

    @pyqtSlot(str)
    def newParamTypeSelected(self, paramName):
        self.requiredTagsListModel.clear()
        self.requiredTagsListModel.refresh()

        paramType  = getParameterTypeFromName(paramName)
        if paramType is None:
            raise ValueError("Parameter type with name '" + paramName + "' was not found.")

        for reqTag in paramType.requiredTags:
            self.requiredTagsListModel.addTag(reqTag.id, reqTag.name, reqTag.id, reqTag.name)

        self.requiredTagsListModel.refresh()


    def buildRequiredTagsGB(self):

        # Widgets            
        self.requireTagGB = QGroupBox("Required tag categories", self)

        self.requiredTagsListTblWdg      = RequiredTagsTableView() 
        self.requiredTagsListModel       = RequiredTagsListModel(parent=self)
        self.requiredTagsListTblWdg.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.requiredTagsListTblWdg.setSelectionMode(QAbstractItemView.SingleSelection)
        self.requiredTagsListTblWdg.setModel(self.requiredTagsListModel)
        self.requiredTagsListTblWdg.setColumnWidth(0, 200)
        self.requiredTagsListTblWdg.setColumnWidth(1, 200)

        # Layout
        requiredTagLayout = QGridLayout(self.requireTagGB)
        requiredTagLayout.addWidget(self.requiredTagsListTblWdg, 0, 0, 4, 1)


    def newParameter(self):
        self.resultTypeCbo.setCurrentIndex(0)
        self.paramModStack.currentWidget().newParameter()
        
        self.singleValueParamWgt.newParameter()
        self.functionParamWgt.newParameter()
        self.traceParamWgt.newParameter()       
        
        self.newParamsGB.setEnabled(True)
        self.paramListTblWdg.clearSelection()

        self.newParamBtn.setEnabled(False)
        self.deleteParamBtn.setEnabled(False)
        self.paramSaveAnnotBtn.setEnabled(True)

        self.isExpProp.setChecked(False)


    def saveParameter(self):

        relationship = self.relationWgt.getRelationship()
        
        # Get the ID of the modified parameter if we are modifying an existing
        # parameters        
        if len(self.paramListTblWdg.selectionModel().selectedRows()) != 0:
            selectedRow = self.paramListTblWdg.selectionModel().currentIndex().row()
            paramId = self.main_window.currentAnnotation.parameters[selectedRow].id
        else:
            paramId = None
        
        param = self.paramModStack.currentWidget().saveParameter(relationship, paramId)

        if not param is None:
            param.requiredTags             = self.requiredTagsListModel.getRequiredTags()
            param.isExperimentProperty     = self.isExpProp.isChecked()

            selectedRow = self.paramListTblWdg.selectionModel().currentIndex().row()
            # Even when there is no selection, selectedRow can take a zero value. This "if" 
            # controls for that.
            if len(self.paramListTblWdg.selectionModel().selectedRows()) == 0:
                selectedRow = -1

            if selectedRow >= 0:
                self.main_window.currentAnnotation.parameters[selectedRow] = param
            else:
                self.main_window.currentAnnotation.parameters.append(param)

            self.additionMode = False
            nbParams = len(self.main_window.currentAnnotation.parameters)
            self.main_window.saveAnnotation()

            if selectedRow < 0 :
                selectedRow = nbParams-1
            self.paramListTblWdg.selectRow(selectedRow)
            self.loadRow(selectedRow)     



    def deleteParameter(self):
        selectedRow = self.paramListTblWdg.selectionModel().currentIndex().row()
        del self.main_window.currentAnnotation.parameters[selectedRow]
        self.main_window.saveAnnotation()
        self.refreshModelingParameters()


    def refreshModelingParameters(self):
        selectedRow = self.paramListTblWdg.selectionModel().currentIndex().row()
        self.loadModelingParameter(selectedRow)


    def loadModelingParameter(self, row = None):
        """
         Call when a new annotation has been selected so that all the modeling parameters
         associated with this annotation are loaded in the parameter list. 
        """

        self.requiredTagsListModel.clear()
        self.requiredTagsListModel.refresh()

        if self.main_window.currentAnnotation is None:
            self.paramListModel.parameterList = []
        else:
            self.paramListModel.parameterList = self.main_window.currentAnnotation.parameters

            aRowIsSelected = not row is None
            if aRowIsSelected:
                if row < 0:
                    noRowToLoad = self.paramListTblWdg.model().rowCount()-row
                else:
                    noRowToLoad = row
            else:
                ## No rows are selected
                noRowToLoad = -1
            
            self.loadRow(noRowToLoad)
            self.newParamBtn.setEnabled(True)
            self.deleteParamBtn.setEnabled(aRowIsSelected)
            self.paramSaveAnnotBtn.setEnabled(aRowIsSelected)
            self.paramModStack.currentWidget().loadModelingParameter(row)
            self.relationWgt.loadModelingParameter(row)

            self.newParamsGB.setEnabled(aRowIsSelected)
        
        self.paramListModel.refresh()


    def loadRow(self, selectedRow = None):
        """
         Called when a row has been selected in the table listing all the modeling parameters.
         It update the interface with the values associated with this specific parameter.
        """
        
        def nlxCheck(id):
            if id in nlx2ks:
                return nlx2ks[id]
            return id
            
        def clear():
            self.requiredTagsListModel.clear()
            self.paramModStack.currentWidget().loadRow(None)
            self.relationWgt.clear()        
            self.paramListTblWdg.clearSelection()

        if selectedRow is None:
            selectedRow = self.paramListTblWdg.selectionModel().currentIndex().row()
        
        if self.main_window.currentAnnotation is None:
            clear()
            return
        
        if selectedRow < 0 or selectedRow >= len(self.main_window.currentAnnotation.parameters) :
            clear()
            return
            
        currentParameter = self.main_window.currentAnnotation.parameters[selectedRow]
    
        self.newParamBtn.setEnabled(True)
        self.deleteParamBtn.setEnabled(True)
        self.paramSaveAnnotBtn.setEnabled(True)

        if currentParameter.description.type == "pointValue":
            self.resultTypeCbo.setCurrentIndex(0)
            self.paramModStack.setCurrentIndex(0)
        elif currentParameter.description.type == "function": 
            self.resultTypeCbo.setCurrentIndex(1)
            self.paramModStack.setCurrentIndex(1)
        elif currentParameter.description.type == "numericalTrace": 
            self.resultTypeCbo.setCurrentIndex(2)
            self.paramModStack.setCurrentIndex(2)
        else:
            raise ValueError("Type of parameter description " + currentParameter.description.type + " is invalid.")

        self.paramModStack.currentWidget().loadRow(currentParameter)
        self.relationWgt.loadRow(currentParameter)
        self.isExpProp.setChecked(currentParameter.isExperimentProperty)

        ## UPDATING REQUIRED TAGS
        self.requiredTagsListModel.clear()       
        for tag in currentParameter.requiredTags:        
            self.requiredTagsListModel.addTag(tag.rootId, self.main_window.dicData[tag.rootId], tag.id, tag.name)
        
        ## Adding new required tags that may have been specified since the 
        ## creation of this parameter instance.
        parameterType = getParameterTypeFromID(currentParameter.typeId)
        reqTags = {reqTag.rootId:reqTag for reqTag in parameterType.requiredTags}
        for reqTagRootId, reqTag in reqTags.items():
            #print(nlxCheck(reqTagRootId), [nlxCheck(tag.rootId) for tag in currentParameter.requiredTags])
            if not nlxCheck(reqTagRootId) in [nlxCheck(tag.rootId) for tag in currentParameter.requiredTags]:
                self.requiredTagsListModel.addTag(reqTag.rootId, self.main_window.dicData[reqTag.rootId], reqTag.id, reqTag.name)
            
        self.requiredTagsListModel.refresh()

        self.newParamsGB.setEnabled(True)


    def selectedParameterChanged(self, selected, deselected):
        if len(selected.indexes()) == 0:
            return
        if self.additionMode:
            msgBox = QMessageBox(self)
            msgBox.setWindowTitle("Cancellation")
            msgBox.setText("Are you sure you want to cancel the addition of the new parameter being edited? If not, say no and then hit 'Save' to save this new parameter.")
            msgBox.setStandardButtons(QMessageBox.No | QMessageBox.Yes)
            msgBox.setDefaultButton(QMessageBox.No)
            if msgBox.exec_() == QMessageBox.Yes:
                self.additionMode = False
                self.loadRow()
            else:
                #self.paramListTblWdg.selectRow(-1)
                self.paramListTblWdg.clearSelection()
        else:
            self.loadRow()


class ParameterListModel(QAbstractTableModel):

    def __init__(self, parameterList = [], header = ['ID', 'Type', 'Description'], parent=None):
        super().__init__(parent)
        self.parameterList = parameterList
        self.header = header
        self.nbCol = len(header)

    def rowCount(self, parent=QModelIndex()):
        return len(self.parameterList)

    def columnCount(self, parent=QModelIndex()):
        return self.nbCol 


    def getSelectedParameter(self, selection):

        if isinstance(selection, list):
            if selection == []:
                return None
            elif isinstance(selection[0], QModelIndex):
                index = selection[0]
        else:
            if selection.at(0) is None:
                return None
            index = selection.at(0).indexes()[0]
        return self.parameterList[index.row()]



    def getByIndex(self, param, ind):
        if ind == 0:
            return param.id
        elif ind == 1:
            return param.typeDesc
        elif ind == 2:
            if param.typeDesc == "pointValue":
                return param.name + " = " + param.description.depVar.values.text(True)
            elif param.typeDesc == "function":
                return param.description.equation
            elif param.typeDesc == "numericalTrace":
                return getParameterTypeNameFromID(param.description.depVar.typeId) \
                          + "=f(" + " ,".join([getParameterTypeNameFromID(v.typeId) for v in param.description.indepVars]) +  ")"
            else:
                raise ValueError
        else:
            raise ValueError

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None

        if role != Qt.DisplayRole:
            return None

        return self.getByIndex(self.parameterList[index.row()], index.column())

    def headerData(self, col, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.header[col]
        return None

    def sort(self, col, order=Qt.AscendingOrder):
        # Sort table by given col number col.
        self.layoutAboutToBeChanged.emit()
        reverse = (order == Qt.DescendingOrder)
        self.annotationList = sorted(self.parameterList, key=lambda x: x.getByIndex(col), reverse = reverse)
        self.layoutChanged.emit()

    def refresh(self):
        self.layoutChanged.emit()



#!/usr/bin/python3

__author__ = "Christian O'Reilly"

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QColor, QPalette
from PyQt5.QtWidgets import QMessageBox


paleGray = QColor(215, 214, 213)

def errorMessage(selfObj, title, message):
    msgBox = QMessageBox(selfObj)
    msgBox.setWindowTitle(title)
    msgBox.setText(message)
    msgBox.exec_()


def disableTextWidget(widget):
    palette = QPalette()
    palette.setColor(QPalette.Base,paleGray)

    widget.setReadOnly(True)
    widget.setPalette(palette)


def enableTextWidget(widget):
    palette = QPalette()
    palette.setColor(QPalette.Base, Qt.white)

    widget.setReadOnly(False)
    widget.setPalette(palette)





__authors__ = ["Pierre-Alexandre Fonta", "Christian O'Reilly"]
__maintainer__ = "Pierre-Alexandre Fonta"

from PyQt5.QtCore import Qt, pyqtSlot, QSortFilterProxyModel
from PyQt5.QtWidgets import (QAbstractItemView, QWidget, QTableView,
                             QVBoxLayout, QLineEdit, QFormLayout)

from nat.zotero_wrap import ZoteroWrap
from neurocurator import utils
from neurocurator.zotero_edition import ZoteroReferenceDialog
from neurocurator.zotero_model import ZoteroTableModel
from neurocurator.zotero_thread import ZoteroRefreshThread


class ZoteroTableWidget(QWidget):

    def __init__(self, settings, directory, check_id_fct, annotations_path, parent=None):
        super().__init__(parent)
        # FIXME Delayed refactoring of check_id_fct and annotations_path.

        # Variables section.

        library_id = settings["libraryID"]
        library_type = settings["libraryType"]
        api_key = settings["apiKey"]
        self._zotero = ZoteroWrap(library_id, library_type, api_key, directory)

        # Widgets section.

        model = ZoteroTableModel(self._zotero, check_id_fct, annotations_path)
        model.load()

        proxy_model = QSortFilterProxyModel()
        proxy_model.setSourceModel(model)
        proxy_model.setDynamicSortFilter(True)
        proxy_model.setFilterCaseSensitivity(Qt.CaseInsensitive)
        proxy_model.setFilterKeyColumn(-1)  # NB: All columns.

        self.view = QTableView(self)
        self.view.setModel(proxy_model)
        self.view.setCornerButtonEnabled(False)
        self.view.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.view.setSelectionMode(QAbstractItemView.SingleSelection)
        # NB: Triggers a call to sortByColumn() which sorts by the first column.
        self.view.setSortingEnabled(True)
        self.view.setWordWrap(False)
        self.view.verticalHeader().hide()

        self.filter_edit = FilterEdit(self.view)

        # NB: The thread does not begin executing until start() is called.
        self.refresh_thread = ZoteroRefreshThread(model, self)

        # Layouts section.

        header_layout = QFormLayout()
        header_layout.addRow("Filter:", self.filter_edit)
        header_layout.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)
        utils.configure_form_layout(header_layout)

        main_layout = QVBoxLayout()
        main_layout.addLayout(header_layout)
        main_layout.addWidget(self.view)
        self.setLayout(main_layout)

        # Signals section.

        self.filter_edit.textChanged.connect(proxy_model.setFilterFixedString)
        self.refresh_thread.started.connect(self.refresh_started)
        self.refresh_thread.finished.connect(self.refresh_finished)

    # def __del__(self):
    #     # FIXME Delayed refactoring. Not called when application is closed. Incorrect parent use?
    #     # NB: Exiting the program when another thread is still busy is a programming error.
    #     # NB: Call QThread::quit() if the thread has an event loop.
    #     print("DEBUG: ZoteroTableWidget.__del__()")
    #     # TODO Display an information dialog.
    #     self.refresh_thread.wait()
    #     print("DEBUG: ZoteroRefreshThread.wait() returned")

    # Slots section.

    @pyqtSlot()
    def refresh_database(self):
        """Start the thread refreshing the Zotero data.

        If the thread is already running, it is not restarted.
        """
        self.refresh_thread.start()

    @pyqtSlot()
    def refresh_started(self):
        """Disable the Zotero widget when the thread refreshing its data runs.

        Disable handling of keyboard/mouse events to ensure a thread-safe refresh.
        """
        # TODO Display an information on top of the disabled widget.
        self.setDisabled(True)

    @pyqtSlot()
    def refresh_finished(self):
        """Enable the Zotero widget when the thread refreshing its data finishes.

        Reset the selection model of the view in case of new/deleted references.
        Enable again the handling of keyboard/mouse events.
        """
        self.view.selectionModel().reset()
        self.setEnabled(True)

    @pyqtSlot()
    def add_reference(self):
        """Display the form for and handle the creation of a new reference."""
        dialog = ZoteroReferenceDialog(self._zotero.reference_templates, self)
        dialog.setWindowTitle("Zotero reference creation")
        dialog.select_reference_type("journalArticle")
        # NB: exec() always pops up the dialog as modal.
        if dialog.exec():
            reference_data = dialog.reference_data()
            # FIXME DEBUG.
            # print("\n")
            # print("REFERENCE DATA: " + repr(reference_data))
            # /FIXME DEBUG.
            # TODO Implement an offline mode. Catch PyZoteroError.
            reference = self._zotero.create_distant_reference(reference_data)
            source_index = self.view.model().sourceModel().add_reference(reference)
            proxy_index = self.view.model().mapFromSource(source_index)
            self.view.selectRow(proxy_index.row())

    @pyqtSlot()
    def edit_reference(self):
        """Display the form for and handle the edition of the selected reference."""
        selected = self.view.selectionModel().currentIndex()
        if selected.isValid():
            row = self.view.model().mapToSource(selected).row()
            # FIXME DEBUG.
            # print("\n")
            # print("TITLE: " + self.zotero.reference_title(row))
            # print("SOURCE ROW NB: " + str(row))
            # /FIXME DEBUG.
            reference_key = self._zotero.reference_key(row)
            # TODO Refresh local when distant is different and 'Cancel' clicked.
            # TODO Display an information dialog when local and distant are different.
            # TODO Implement an offline mode. Catch PyZoteroError.
            reference = self._zotero.get_reference(reference_key)
            dialog = ZoteroReferenceDialog(self._zotero.reference_templates, self)
            dialog.setWindowTitle("Zotero reference edition")
            dialog.load_reference_data(reference["data"])
            # NB: exec() always pops up the dialog as modal.
            if dialog.exec():
                reference["data"] = dialog.reference_data()
                # FIXME DEBUG.
                # print("\n")
                # print("REFERENCE DATA: " + repr(reference["data"]))
                # /FIXME DEBUG.
                # TODO Implement an offline mode. Catch PyZoteroError.
                self._zotero.update_distant_reference(reference)
                # TODO Implement an offline mode. Catch PyZoteroError.
                reference = self._zotero.get_reference(reference_key)
                self.view.model().sourceModel().update_reference(row, reference)
        else:
            # TODO Display an information dialog.
            raise ValueError("No row selected for edition.")


class FilterEdit(QLineEdit):
    # QTableView::setSelectionMode(QAbstractItemView.NoSelection) without or
    # with QTableView::clearSelection() is not sufficient to disable the
    # selection of a row during QSortFilterProxyModel::setFilterFixedString().
    # The purpose of this subclassing is to disable this automatic selection.

    def __init__(self, view, parent=None):
        super().__init__(parent)
        self._view = view

    # Qt interface implementation section.

    def focusInEvent(self, event):
        # QItemSelectionModel::clearSelection() is not sufficient.
        # QItemSelectionModel::clearSelection() emits signals, reset() not.
        self._view.selectionModel().reset()
        super().focusInEvent(event)
